"""
Components - ChrisLR 2022

This file contains the Descriptors used to set Fields in Components
"""
import typing

from evennia.typeclasses.attributes import AttributeProperty, NAttributeProperty

if typing.TYPE_CHECKING:
    from .components import Component


class DBField(AttributeProperty):
    """
    Component Attribute Descriptor.
    Allows you to set attributes related to a component on the class.
    It uses AttributeProperty under the hood but prefixes the key with the component name.
    """

    def __init__(self, default=None, autocreate=False, **kwargs):
        super().__init__(default=default, autocreate=autocreate, **kwargs)

    def __set_name__(self, owner: "Component", name):
        """
        Called when descriptor is first assigned to the class.

        Args:
            owner (Component): The component classF on which this is set
            name (str): The name that was used to set the DBField.
        """
        self._key = f"{owner.get_component_slot()}::{name}"
        owner.add_field(name, self)

    def at_added(self, component):
        """
        Called when the parent component is added to a host.

        Args:
            component (Component): The component instance being added.
        """

        if self._autocreate:
            self.__get__(component, type(component))

    def at_removed(self, component):
        """
        Called when the parent component is removed from a host.

        Args:
            component (Component): The component instance being removed.
        """

        self.__delete__(component)


class NDBField(NAttributeProperty):
    """
    Component In-Memory Attribute Descriptor.
    Allows you to set in-memory attributes related to a component on the class.
    It uses NAttributeProperty under the hood but prefixes the key with the component name.
    """

    def __set_name__(self, owner: "Component", name):
        """
        Called when descriptor is first assigned to the class.

        Args:
            owner (Component): The component class on which this is set
            name (str): The name that was used to set the DBField.
        """
        self._key = f"{owner.get_component_slot()}::{name}"
        owner.add_field(name, self)

    def at_added(self, component):
        """
        Called when the parent component is added to a host.

        Args:
            component (Component): The component instance being added.
        """
        if self._autocreate:
            self.__set__(component, self._default)

    def at_removed(self, component):
        """
        Called when the parent component is removed from a host.

        Args:
            component (Component): The component instance being removed.
        """
        self.__delete__(component)


class TagField:
    """
    Component Tags Descriptor.
    Allows you to set Tags related to a component on the class.
    The tags are set with a prefixed category, so it can support
    multiple tags or enforce a single one.

    Default value of a tag is added when the component is registered.
    Tags are removed if the component itself is removed.
    """

    def __init__(self, default=None, enforce_single=False):
        self._category_key = None
        self._default = default
        self._enforce_single = enforce_single

    def __set_name__(self, owner: "Component", name):
        """
        Called when TagField is first assigned to the class.
        It is called with the component class and the name of the field.
        """
        self._category_key = f"{owner.get_component_slot()}::{name}"
        owner.add_field(name, self)

    def __get__(self, instance, owner):
        """
        Called when retrieving the value of the TagField.
        It is called with the component instance and the class.
        """
        tag_value = instance.host.tags.get(
            default=self._default,
            category=self._category_key,
        )
        return tag_value

    def __set__(self, instance, value):
        """
        Called when setting a value on the TagField.
        It is called with the component instance and the value.
        """

        tag_handler = instance.host.tags
        if self._enforce_single:
            tag_handler.clear(category=self._category_key)

        tag_handler.add(
            key=value,
            category=self._category_key,
        )

    def __delete__(self, instance):
        """
        Used when 'del' is called on the TagField.
        It is called with the component instance.
        """
        instance.host.tags.clear(category=self._category_key)

    def at_added(self, component):
        """
        Called when the parent component is added to a host.

        Args:
            component (Component): The component instance being added.
        """
        if self._default:
            self.__set__(component, self._default)

    def at_removed(self, component):
        """
        Called when the parent component is removed from a host.

        Args:
            component (Component): The component instance being removed.
        """
        self.__delete__(component)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.11 on 2017-01-25 22:30


from django.db import migrations


def update_tags_with_dbmodel(apps, schema_editor):

    Tag = apps.get_model("typeclasses", "Tag")

    ObjectDB = apps.get_model("objects", "ObjectDB")
    for obj in ObjectDB.objects.all().exclude(db_tags__db_model="objectdb"):
        for tag in obj.db_tags.all().exclude(db_model="objectdb"):
            obj.db_tags.remove(tag)
            newtag = Tag(
                db_key=tag.db_key,
                db_category=tag.db_category,
                db_data=tag.db_data,
                db_tagtype=tag.db_tagtype,
                db_model="objectdb",
            )
            newtag.save()
            obj.db_tags.add(newtag)

    AccountDB = apps.get_model("accounts", "AccountDB")
    for obj in AccountDB.objects.all().exclude(db_tags__db_model="accountdb"):
        for tag in obj.db_tags.all().exclude(db_model="accountdb"):
            obj.db_tags.remove(tag)
            newtag = Tag(
                db_key=tag.db_key,
                db_category=tag.db_category,
                db_data=tag.db_data,
                db_tagtype=tag.db_tagtype,
                db_model="accountdb",
            )
            newtag.save()
            obj.db_tags.add(newtag)

    ScriptDB = apps.get_model("scripts", "ScriptDB")
    for obj in ScriptDB.objects.all().exclude(db_tags__db_model="scriptdb"):
        for tag in obj.db_tags.all().exclude(db_model="scriptdb"):
            obj.db_tags.remove(tag)
            newtag = Tag(
                db_key=tag.db_key,
                db_category=tag.db_category,
                db_data=tag.db_data,
                db_tagtype=tag.db_tagtype,
                db_model="scriptdb",
            )
            newtag.save()
            obj.db_tags.add(newtag)

    ChannelDB = apps.get_model("comms", "ChannelDB")
    for obj in ChannelDB.objects.all().exclude(db_tags__db_model="channeldb"):
        for tag in obj.db_tags.all().exclude(db_model="channeldb"):
            obj.db_tags.remove(tag)
            newtag = Tag(
                db_key=tag.db_key,
                db_category=tag.db_category,
                db_data=tag.db_data,
                db_tagtype=tag.db_tagtype,
                db_model="channeldb",
            )
            newtag.save()
            obj.db_tags.add(newtag)

    HelpEntry = apps.get_model("help", "HelpEntry")
    for obj in HelpEntry.objects.all().exclude(db_tags__db_model="helpentry"):
        for tag in obj.db_tags.all().exclude(db_model="helpentry"):
            obj.db_tags.remove(tag)
            newtag = Tag(
                db_key=tag.db_key,
                db_category=tag.db_category,
                db_data=tag.db_data,
                db_tagtype=tag.db_tagtype,
                db_model="helpentry",
            )
            newtag.save()
            obj.db_tags.add(newtag)

    Msg = apps.get_model("comms", "Msg")
    for obj in Msg.objects.all().exclude(db_tags__db_model="msg"):
        for tag in obj.db_tags.all().exclude(db_model="msg"):
            obj.db_tags.remove(tag)
            newtag = Tag(
                db_key=tag.db_key,
                db_category=tag.db_category,
                db_data=tag.db_data,
                db_tagtype=tag.db_tagtype,
                db_model="msg",
            )
            newtag.save()
            obj.db_tags.add(newtag)


class Migration(migrations.Migration):

    dependencies = [("typeclasses", "0006_auto_add_dbmodel_value_for_tags_attributes")]

    operations = [
        migrations.AlterUniqueTogether(
            name="tag", unique_together=set([("db_key", "db_category", "db_tagtype", "db_model")])
        ),
        migrations.AlterIndexTogether(
            name="tag", index_together=set([("db_key", "db_category", "db_tagtype", "db_model")])
        ),
        migrations.RunPython(update_tags_with_dbmodel),
    ]



"""
All items in the game inherit from a base object. The properties (what you can do
with an object, such as wear, wield, eat, drink, kill etc) are all controlled by
Tags.

Every object has one of a few `obj_type`-category tags:
- weapon
- armor
- shield
- helmet
- consumable  (potions, torches etc)
- magic (runestones, magic items)
- quest (quest-items)
- treasure  (valuable to sell)

It's possible for an item to have more than one tag, such as a golden helmet (helmet+treasure) or
rune sword (weapon+quest).

"""

from evennia import AttributeProperty, create_object, search_object
from evennia.objects.objects import DefaultObject
from evennia.utils.utils import make_iter

from . import rules
from .enums import Ability, ObjType, WieldLocation
from .utils import get_obj_stats

_BARE_HANDS = None


class EvAdventureObject(DefaultObject):
    """
    Base in-game entity.

    """

    # inventory management
    inventory_use_slot = AttributeProperty(WieldLocation.BACKPACK)
    # how many inventory slots it uses (can be a fraction)
    size = AttributeProperty(1)
    value = AttributeProperty(0)

    # can also be an iterable, for adding multiple obj-type tags
    obj_type = ObjType.GEAR

    def at_object_creation(self):
        for obj_type in make_iter(self.obj_type):
            self.tags.add(obj_type.value, category="obj_type")

    def get_display_header(self, looker, **kwargs):
        return ""  # this is handled by get_obj_stats

    def get_display_desc(self, looker, **kwargs):
        return get_obj_stats(self, owner=looker)

    def has_obj_type(self, objtype):
        """
        Check if object is of a particular type.

        typeobj_enum (enum.ObjType): A type to check, like enums.TypeObj.TREASURE.

        """
        return objtype.value in make_iter(self.obj_type)

    def get_help(self):
        """
        Get help text for the item.

        Returns:
            str: The help text, by default taken from the `.help_text` property.

        """
        return "No help for this item."

    def at_pre_use(self, *args, **kwargs):
        """
        Called before use. If returning False, usage should be aborted.
        """
        return True

    def use(self, *args, **kwargs):
        """
        Use this object, whatever that may mean.

        """
        raise NotImplementedError

    def at_post_use(self, *args, **kwargs):
        """
        Called after use happened.
        """
        pass


class EvAdventureObjectFiller(EvAdventureObject):
    """
    In _Knave_, the inventory slots act as an extra measure of how you are affected by
    various averse effects. For example, mud or water could fill up some of your inventory
    slots and make the equipment there unusable until you cleaned it. Inventory is also
    used to track how long you can stay under water etc - the fewer empty slots you have,
    the less time you can stay under water due to carrying so much stuff with you.

    This class represents such an effect filling up an empty slot. It has a quality of 0,
    meaning it's unusable.

    """

    obj_type = ObjType.QUEST.value  # can't be sold
    quality = AttributeProperty(0)


class EvAdventureQuestObject(EvAdventureObject):
    """
    A quest object. These cannot be sold and only be used for quest resolution.

    """

    obj_type = ObjType.QUEST
    value = AttributeProperty(0)


class EvAdventureTreasure(EvAdventureObject):
    """
    A 'treasure' is mainly useful to sell for coin.

    """

    obj_type = ObjType.TREASURE
    value = AttributeProperty(100, autocreate=False)


class EvAdventureConsumable(EvAdventureObject):
    """
    Item that can be 'used up', like a potion or food. Weapons, armor etc does not
    have a limited usage in this way.

    """

    obj_type = ObjType.CONSUMABLE
    size = AttributeProperty(0.25, autocreate=False)
    uses = AttributeProperty(1, autocreate=False)

    def at_pre_use(self, user, target=None, *args, **kwargs):
        if target and user.location != target.location:
            user.msg("You are not close enough to the target!")
            return False

        if self.uses <= 0:
            user.msg(f"|w{self.key} is used up.|n")
            return False

        return super().at_pre_use(user, target=target, *args, **kwargs)

    def use(self, user, target=None, *args, **kwargs):
        """
        Use the consumable.

        """

        if user.location:
            user.location.msg_contents(
                f"$You() $conj(use) {self.get_display_name(user)}.", from_obj=user
            )

    def at_post_use(self, user, *args, **kwargs):
        """
        Called after this item was used.

        Args:
            user (Object): The one using the item.
            *args, **kwargs: Optional arguments.

        """
        self.uses -= 1
        if self.uses <= 0:
            user.msg(f"|w{self.key} was used up.|n")
            self.delete()


class EvAdventureWeapon(EvAdventureObject):
    """
    Base weapon class for all EvAdventure weapons.

    """

    obj_type = ObjType.WEAPON
    inventory_use_slot = AttributeProperty(WieldLocation.WEAPON_HAND)
    quality = AttributeProperty(3)

    # what ability used to attack with this weapon
    attack_type = AttributeProperty(Ability.STR)
    # what defense stat of the enemy it must defeat
    defense_type = AttributeProperty(Ability.ARMOR)
    damage_roll = AttributeProperty("1d6")

    def get_display_name(self, looker=None, **kwargs):
        quality = self.quality

        quality_txt = ""
        if quality <= 0:
            quality_txt = "|r(broken!)|n"
        elif quality < 2:
            quality_txt = "|y(damaged)|n"
        elif quality < 3:
            quality_txt = "|Y(chipped)|n"

        return super().get_display_name(looker=looker, **kwargs) + quality_txt

    def at_pre_use(self, user, target=None, *args, **kwargs):
        if target and user.location != target.location:
            # we assume weapons can only be used in the same location
            user.msg("You are not close enough to the target!")
            return False

        if self.quality is not None and self.quality <= 0:
            user.msg(f"{self.get_display_name(user)} is broken and can't be used!")
            return False
        return super().at_pre_use(user, target=target, *args, **kwargs)

    def use(self, attacker, target, *args, advantage=False, disadvantage=False, **kwargs):
        """When a weapon is used, it attacks an opponent"""

        location = attacker.location

        is_hit, quality, txt = rules.dice.opposed_saving_throw(
            attacker,
            target,
            attack_type=self.attack_type,
            defense_type=self.defense_type,
            advantage=advantage,
            disadvantage=disadvantage,
        )
        location.msg_contents(
            f"$You() $conj(attack) $You({target.key}) with {self.key}: {txt}",
            from_obj=attacker,
            mapping={target.key: target},
        )
        if is_hit:
            # enemy hit, calculate damage
            dmg = rules.dice.roll(self.damage_roll)

            if quality is Ability.CRITICAL_SUCCESS:
                # doble damage roll for critical success
                dmg += rules.dice.roll(self.damage_roll)
                message = (
                    f" $You() |ycritically|n $conj(hit) $You({target.key}) for |r{dmg}|n damage!"
                )
            else:
                message = f" $You() $conj(hit) $You({target.key}) for |r{dmg}|n damage!"

            location.msg_contents(message, from_obj=attacker, mapping={target.key: target})
            # call hook
            target.at_damage(dmg, attacker=attacker)

        else:
            # a miss
            message = f" $You() $conj(miss) $You({target.key})."
            if quality is Ability.CRITICAL_FAILURE:
                message += ".. it's a |rcritical miss!|n, damaging the weapon."
                if self.quality is not None:
                    self.quality -= 1
                location.msg_contents(message, from_obj=attacker, mapping={target.key: target})

    def at_post_use(self, user, *args, **kwargs):
        if self.quality is not None and self.quality <= 0:
            user.msg(f"|r{self.get_display_name(user)} breaks and can no longer be used!")


class EvAdventureThrowable(EvAdventureWeapon, EvAdventureConsumable):
    """
    Something you can throw at an enemy to harm them once, like a knife or exploding potion/grenade.

    Note: In Knave, ranged attacks are done with WIS (representing the stillness of your mind?)

    """

    obj_type = (ObjType.THROWABLE, ObjType.WEAPON, ObjType.CONSUMABLE)

    attack_type = AttributeProperty(Ability.WIS)
    defense_type = AttributeProperty(Ability.DEX)
    damage_roll = AttributeProperty("1d6")


class EvAdventureRunestone(EvAdventureWeapon, EvAdventureConsumable):
    """
    Base class for magic runestones. In _Knave_, every spell is represented by a rune stone
    that takes up an inventory slot. It is wielded as a weapon in order to create the specific
    magical effect provided by the stone. Normally each stone can only be used once per day but
    they are quite powerful (and scales with caster level).

    """

    obj_type = (ObjType.WEAPON, ObjType.MAGIC)
    inventory_use_slot = WieldLocation.TWO_HANDS
    quality = AttributeProperty(3)

    attack_type = AttributeProperty(Ability.INT)
    defense_type = AttributeProperty(Ability.DEX)
    damage_roll = AttributeProperty("1d8")

    def at_post_use(self, user, *args, **kwargs):
        """Called after the spell was cast"""
        self.uses -= 1
        # the rune stone is not deleted after use, but
        # it needs to be refreshed after resting.

    def refresh(self):
        self.uses = 1


class EvAdventureArmor(EvAdventureObject):
    """
    Base class for all wearable Armors.

    """

    obj_type = ObjType.ARMOR
    inventory_use_slot = WieldLocation.BODY

    armor = AttributeProperty(1)
    quality = AttributeProperty(3)


class EvAdventureShield(EvAdventureArmor):
    """
    Base class for all Shields.

    """

    obj_type = ObjType.SHIELD
    inventory_use_slot = WieldLocation.SHIELD_HAND


class EvAdventureHelmet(EvAdventureArmor):
    """
    Base class for all Helmets.

    """

    obj_type = ObjType.HELMET
    inventory_use_slot = WieldLocation.HEAD


class WeaponBareHands(EvAdventureWeapon):
    """
    This is a dummy-class loaded when you wield no weapons. We won't create any db-object for it.

    """

    obj_type = ObjType.WEAPON
    key = "Bare hands"
    inventory_use_slot = WieldLocation.WEAPON_HAND
    attack_type = Ability.STR
    defense_type = Ability.ARMOR
    damage_roll = "1d4"
    quality = None  # let's assume fists are always available ...


def get_bare_hands():
    """
    Get the bare-hands singleton object.

    Returns:
        WeaponBareHands
    """
    global _BARE_HANDS

    if not _BARE_HANDS:
        _BARE_HANDS = search_object("Bare hands", typeclass=WeaponBareHands).first()
    if not _BARE_HANDS:
        _BARE_HANDS = create_object(WeaponBareHands, key="Bare hands")
    return _BARE_HANDS



# -*- coding: utf-8 -*-


from django.db import migrations, models

import evennia.accounts.manager


class Migration(migrations.Migration):

    dependencies = [
        ("objects", "0005_auto_20150403_2339"),
        ("scripts", "0007_auto_20150403_2339"),
        ("accounts", "0004_auto_20150403_2339"),
        ("typeclasses", "0002_auto_20150109_0913"),
    ]

    operations = [
        migrations.CreateModel(
            name="DefaultObject", fields=[], options={"proxy": True}, bases=("objects.objectdb",)
        ),
        migrations.CreateModel(
            name="DefaultAccount",
            fields=[],
            options={"proxy": True},
            bases=("accounts.accountdb",),
            managers=[("objects", evennia.accounts.manager.AccountManager())],
        ),
        migrations.CreateModel(
            name="ScriptBase", fields=[], options={"proxy": True}, bases=("scripts.scriptdb",)
        ),
        migrations.CreateModel(
            name="DefaultCharacter",
            fields=[],
            options={"proxy": True},
            bases=("typeclasses.defaultobject",),
        ),
        migrations.CreateModel(
            name="DefaultExit",
            fields=[],
            options={"proxy": True},
            bases=("typeclasses.defaultobject",),
        ),
        migrations.CreateModel(
            name="DefaultGuest",
            fields=[],
            options={"proxy": True},
            bases=("typeclasses.defaultaccount",),
            managers=[("objects", evennia.accounts.manager.AccountManager())],
        ),
        migrations.CreateModel(
            name="DefaultRoom",
            fields=[],
            options={"proxy": True},
            bases=("typeclasses.defaultobject",),
        ),
        migrations.CreateModel(
            name="DefaultScript",
            fields=[],
            options={"proxy": True},
            bases=("typeclasses.scriptbase",),
        ),
        migrations.CreateModel(
            name="DoNothing",
            fields=[],
            options={"proxy": True},
            bases=("typeclasses.defaultscript",),
        ),
        migrations.CreateModel(
            name="Store", fields=[], options={"proxy": True}, bases=("typeclasses.defaultscript",)
        ),
    ]



try:
    from django.utils.unittest import TestCase
except ImportError:
    from django.test import TestCase

try:
    from django.utils import unittest
except ImportError:
    import unittest

import json
import pickle
import string
import sys

import mock
from autobahn.twisted.websocket import WebSocketServerFactory
from mock import MagicMock, Mock
from twisted.conch.telnet import DO, DONT, IAC, NAWS, SB, SE, WILL
from twisted.internet.base import DelayedCall
from twisted.test import proto_helpers
from twisted.trial.unittest import TestCase as TwistedTestCase

import evennia
from evennia.server.portal import irc
from evennia.server.portal.portalsessionhandler import PortalSessionHandler
from evennia.server.portal.service import EvenniaPortalService
from evennia.utils.test_resources import BaseEvenniaTest

from .amp import (
    AMP_MAXLEN,
    AMPMultiConnectionProtocol,
    MsgPortal2Server,
    MsgServer2Portal,
)
from .amp_server import AMPServerFactory
from .mccp import MCCP
from .mssp import MSSP
from .mxp import MXP
from .naws import DEFAULT_HEIGHT, DEFAULT_WIDTH
from .suppress_ga import SUPPRESS_GA
from .telnet import TelnetProtocol, TelnetServerFactory
from .telnet_oob import MSDP, MSDP_VAL, MSDP_VAR
from .ttype import IS, TTYPE
from .webclient import WebSocketClient


class TestAMPServer(TwistedTestCase):
    """
    Test AMP communication
    """

    def setUp(self):
        super().setUp()
        portal = Mock()
        factory = AMPServerFactory(portal)
        self.proto = factory.buildProtocol(("localhost", 0))
        self.transport = MagicMock()  # proto_helpers.StringTransport()
        self.transport.client = ["localhost"]
        self.transport.write = MagicMock()

    def test_amp_out(self):
        self.proto.makeConnection(self.transport)

        self.proto.data_to_server(MsgServer2Portal, 1, test=2)

        if pickle.HIGHEST_PROTOCOL == 5:
            # Python 3.8+
            byte_out = (
                b"\x00\x04_ask\x00\x011\x00\x08_command\x00\x10MsgServer2Portal\x00\x0b"
                b"packed_data\x00 x\xdak`\x9d*\xc8\x00\x01\xde\x8c\xb5SzXJR"
                b"\x8bK\xa6x3\x15\xb7M\xd1\x03\x00VU\x07u\x00\x00"
            )
        elif pickle.HIGHEST_PROTOCOL == 4:
            # Python 3.7
            byte_out = (
                b"\x00\x04_ask\x00\x011\x00\x08_command\x00\x10MsgServer2Portal\x00\x0b"
                b"packed_data\x00 x\xdak`\x99*\xc8\x00\x01\xde\x8c\xb5SzXJR"
                b"\x8bK\xa6x3\x15\xb7M\xd1\x03\x00V:\x07t\x00\x00"
            )
        self.transport.write.assert_called_with(byte_out)
        with mock.patch("evennia.server.portal.amp.amp.AMP.dataReceived") as mocked_amprecv:
            self.proto.dataReceived(byte_out)
            mocked_amprecv.assert_called_with(byte_out)

    def test_amp_in(self):
        self.proto.makeConnection(self.transport)

        self.proto.data_to_server(MsgPortal2Server, 1, test=2)
        if pickle.HIGHEST_PROTOCOL == 5:
            # Python 3.8+
            byte_out = (
                b"\x00\x04_ask\x00\x011\x00\x08_command\x00\x10MsgPortal2Server\x00\x0b"
                b"packed_data\x00 x\xdak`\x9d*\xc8\x00\x01\xde\x8c\xb5SzXJR"
                b"\x8bK\xa6x3\x15\xb7M\xd1\x03\x00VU\x07u\x00\x00"
            )
        elif pickle.HIGHEST_PROTOCOL == 4:
            # Python 3.7
            byte_out = (
                b"\x00\x04_ask\x00\x011\x00\x08_command\x00\x10MsgPortal2Server\x00\x0b"
                b"packed_data\x00 x\xdak`\x99*\xc8\x00\x01\xde\x8c\xb5SzXJR"
                b"\x8bK\xa6x3\x15\xb7M\xd1\x03\x00V:\x07t\x00\x00"
            )
        self.transport.write.assert_called_with(byte_out)
        with mock.patch("evennia.server.portal.amp.amp.AMP.dataReceived") as mocked_amprecv:
            self.proto.dataReceived(byte_out)
            mocked_amprecv.assert_called_with(byte_out)

    def test_large_msg(self):
        """
        Send message larger than AMP_MAXLEN - should be split into several
        """
        self.proto.makeConnection(self.transport)
        outstr = "test" * AMP_MAXLEN
        self.proto.data_to_server(MsgServer2Portal, 1, test=outstr)

        if pickle.HIGHEST_PROTOCOL == 5:
            # Python 3.8+
            self.transport.write.assert_called_with(
                b"\x00\x04_ask\x00\x011\x00\x08_command\x00\x10MsgServer2Portal\x00\x0bpacked_data"
                b"\x00wx\xda\xed\xc6\xc1\t\x80 \x00@Q#=5Z\x0b\xb8\x80\x13\xe85h\x80\x8e\xbam`Dc\xf4><\xf8g"
                b"\x1a[\xf8\xda\x97\xa3_\xb1\x95\xdaz\xbe\xe7\x1a\xde\x03\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xe0\x1f\x1eP\x1d\x02\r\x00\rpacked_data.2"
                b"\x00Zx\xda\xed\xc3\x01\r\x00\x00\x08\xc0\xa0\xb4&\xf0\xfdg\x10a\xa3"
                b"\xd9RUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU\xf5\xfb\x03m\xe0\x06"
                b"\x1d\x00\rpacked_data.3\x00Zx\xda\xed\xc3\x01\r\x00\x00\x08\xc0\xa0\xb4&\xf0\xfdg\x10a"
                b"\xa3fSUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU\xf5\xfb\x03n\x1c"
                b"\x06\x1e\x00\rpacked_data.4\x00Zx\xda\xed\xc3\x01\t\x00\x00\x0c\x03\xa0\xb4O\xb0\xf5gA"
                b"\xae`\xda\x8b\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa"
                b"\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa"
                b"\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa"
                b"\xaa\xaa\xaa\xdf\x0fnI\x06,\x00\rpacked_data.5\x00\x18x\xdaK-.)I\xc5\x8e\xa7\xb22@\xc0"
                b"\x94\xe2\xb6)z\x00Z\x1e\x0e\xb6\x00\x00"
            )
        elif pickle.HIGHEST_PROTOCOL == 4:
            # Python 3.7
            self.transport.write.assert_called_with(
                b"\x00\x04_ask\x00\x011\x00\x08_command\x00\x10MsgServer2Portal\x00\x0bpacked_data"
                b"\x00wx\xda\xed\xc6\xc1\t\x80 \x00@Q#o\x8e\xd6\x02-\xe0\x04z\r\x1a\xa0\xa3m+$\xd2"
                b"\x18\xbe\x0f\x0f\xfe\x1d\xdf\x14\xfe\x8e\xedjO\xac\xb9\xd4v\xf6o\x0f\xf3\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00X\xc3\x00P\x10\x02\x0c\x00\rpacked_data.2\x00Zx\xda\xed\xc3\x01\r\x00\x00\x08"
                b"\xc0\xa0\xb4&\xf0\xfdg\x10a\xa3\xd9RUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU"
                b"\xf5\xfb\x03m\xe0\x06\x1d\x00\rpacked_data.3\x00Zx\xda\xed\xc3\x01\r\x00\x00\x08"
                b"\xc0\xa0\xb4&\xf0\xfdg\x10a\xa3fSUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU"
                b"\xf5\xfb\x03n\x1c\x06\x1e\x00\rpacked_data.4\x00Zx\xda\xed\xc3\x01\t\x00\x00\x0c"
                b"\x03\xa0\xb4O\xb0\xf5gA\xae`\xda\x8b\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa"
                b"\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa"
                b"\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa"
                b"\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xaa\xdf\x0fnI\x06,\x00\rpacked_data.5"
                b"\x00\x18x\xdaK-.)I\xc5\x8e\xa7\xb22@\xc0\x94\xe2\xb6)z\x00Z\x1e\x0e\xb6\x00\x00"
            )


class TestIRC(TestCase):
    def test_plain_ansi(self):
        """
        Test that printable characters do not get mangled.
        """
        irc_ansi = irc.parse_ansi_to_irc(string.printable)
        ansi_irc = irc.parse_irc_to_ansi(string.printable)
        self.assertEqual(irc_ansi, string.printable)
        self.assertEqual(ansi_irc, string.printable)

    def test_bold(self):
        s_irc = "\x02thisisatest"
        s_eve = r"|hthisisatest"
        self.assertEqual(irc.parse_ansi_to_irc(s_eve), s_irc)
        self.assertEqual(s_eve, irc.parse_irc_to_ansi(s_irc))

    def test_italic(self):
        s_irc = "\x02thisisatest"
        s_eve = r"|hthisisatest"
        self.assertEqual(irc.parse_ansi_to_irc(s_eve), s_irc)

    def test_colors(self):
        color_map = (
            ("\0030", r"|w"),
            ("\0031", r"|X"),
            ("\0032", r"|B"),
            ("\0033", r"|G"),
            ("\0034", r"|r"),
            ("\0035", r"|R"),
            ("\0036", r"|M"),
            ("\0037", r"|Y"),
            ("\0038", r"|y"),
            ("\0039", r"|g"),
            ("\00310", r"|C"),
            ("\00311", r"|c"),
            ("\00312", r"|b"),
            ("\00313", r"|m"),
            ("\00314", r"|x"),
            ("\00315", r"|W"),
            ("\00399,5", r"|[r"),
            ("\00399,3", r"|[g"),
            ("\00399,7", r"|[y"),
            ("\00399,2", r"|[b"),
            ("\00399,6", r"|[m"),
            ("\00399,10", r"|[c"),
            ("\00399,15", r"|[w"),
            ("\00399,1", r"|[x"),
        )

        for m in color_map:
            self.assertEqual(irc.parse_irc_to_ansi(m[0]), m[1])
            self.assertEqual(m[0], irc.parse_ansi_to_irc(m[1]))

    def test_identity(self):
        """
        Test that the composition of the function and
        its inverse gives the correct string.
        """

        s = r"|wthis|Xis|gis|Ma|C|complex|*string"

        self.assertEqual(irc.parse_irc_to_ansi(irc.parse_ansi_to_irc(s)), s)


class TestTelnet(TwistedTestCase):
    def setUp(self):
        super().setUp()
        self.portal = EvenniaPortalService()
        evennia.EVENNIA_PORTAL_SERVICE = self.portal
        self.amp_server_factory = AMPServerFactory(self.portal)
        self.amp_server = self.amp_server_factory.buildProtocol("127.0.0.1")
        factory = TelnetServerFactory()
        factory.protocol = TelnetProtocol
        evennia.PORTAL_SESSION_HANDLER = PortalSessionHandler()
        factory.sessionhandler = evennia.PORTAL_SESSION_HANDLER
        factory.sessionhandler.portal = Mock()
        self.proto = factory.buildProtocol(("localhost", 0))
        self.transport = proto_helpers.StringTransport()
        self.addCleanup(factory.sessionhandler.disconnect_all)

    @mock.patch("evennia.server.portal.portalsessionhandler.reactor", new=MagicMock())
    def test_mudlet_ttype(self):
        self.transport.client = ["localhost"]
        self.transport.setTcpKeepAlive = Mock()
        d = self.proto.makeConnection(self.transport)
        # test suppress_ga
        self.assertTrue(self.proto.protocol_flags["NOGOAHEAD"])
        self.proto.dataReceived(IAC + DONT + SUPPRESS_GA)
        self.assertFalse(self.proto.protocol_flags["NOGOAHEAD"])
        self.assertEqual(self.proto.handshakes, 7)
        # test naws
        self.assertEqual(self.proto.protocol_flags["SCREENWIDTH"], {0: DEFAULT_WIDTH})
        self.assertEqual(self.proto.protocol_flags["SCREENHEIGHT"], {0: DEFAULT_HEIGHT})
        self.proto.dataReceived(IAC + WILL + NAWS)
        self.proto.dataReceived(b"".join([IAC, SB, NAWS, b"", b"x", b"", b"d", IAC, SE]))
        self.assertEqual(self.proto.protocol_flags["SCREENWIDTH"][0], 78)
        self.assertEqual(self.proto.protocol_flags["SCREENHEIGHT"][0], 45)
        self.assertEqual(self.proto.handshakes, 6)
        # test ttype
        self.assertFalse(self.proto.protocol_flags["TTYPE"])
        self.assertTrue(self.proto.protocol_flags["ANSI"])
        self.proto.dataReceived(IAC + WILL + TTYPE)
        self.proto.dataReceived(b"".join([IAC, SB, TTYPE, IS, b"MUDLET", IAC, SE]))
        self.assertTrue(self.proto.protocol_flags["XTERM256"])
        self.assertEqual(self.proto.protocol_flags["CLIENTNAME"], "MUDLET")
        self.assertTrue(self.proto.protocol_flags["FORCEDENDLINE"])
        self.assertTrue(self.proto.protocol_flags["NOGOAHEAD"])
        self.assertFalse(self.proto.protocol_flags["NOPROMPTGOAHEAD"])
        self.proto.dataReceived(b"".join([IAC, SB, TTYPE, IS, b"XTERM", IAC, SE]))
        self.proto.dataReceived(b"".join([IAC, SB, TTYPE, IS, b"MTTS 137", IAC, SE]))
        self.assertEqual(self.proto.handshakes, 5)
        # test mccp
        self.proto.dataReceived(IAC + DONT + MCCP)
        self.assertFalse(self.proto.protocol_flags["MCCP"])
        self.assertEqual(self.proto.handshakes, 4)
        # test mssp
        self.proto.dataReceived(IAC + DONT + MSSP)
        self.assertEqual(self.proto.handshakes, 3)
        # test oob
        self.proto.dataReceived(IAC + DO + MSDP)
        self.proto.dataReceived(
            b"".join([IAC, SB, MSDP, MSDP_VAR, b"LIST", MSDP_VAL, b"COMMANDS", IAC, SE])
        )
        self.assertTrue(self.proto.protocol_flags["OOB"])
        self.assertEqual(self.proto.handshakes, 2)
        # test mxp
        self.proto.dataReceived(IAC + DONT + MXP)
        self.assertFalse(self.proto.protocol_flags["MXP"])
        self.assertEqual(self.proto.handshakes, 1)
        # clean up to prevent Unclean reactor
        self.proto.nop_keep_alive.stop()
        self.proto._handshake_delay.cancel()
        return d


class TestWebSocket(BaseEvenniaTest):
    def setUp(self):
        super().setUp()
        self.portal = EvenniaPortalService()
        evennia.EVENNIA_PORTAL_SERVICE = self.portal
        self.amp_server_factory = AMPServerFactory(self.portal)
        self.amp_server = self.amp_server_factory.buildProtocol("127.0.0.1")
        self.proto = WebSocketClient()
        self.proto.factory = WebSocketServerFactory()
        evennia.PORTAL_SESSION_HANDLER = PortalSessionHandler()
        self.proto.factory.sessionhandler = evennia.PORTAL_SESSION_HANDLER
        self.proto.sessionhandler = evennia.PORTAL_SESSION_HANDLER
        self.proto.sessionhandler.portal = Mock()
        self.proto.transport = proto_helpers.StringTransport()
        # self.proto.transport = proto_helpers.FakeDatagramTransport()
        self.proto.transport.client = ["localhost"]
        self.proto.transport.setTcpKeepAlive = Mock()
        self.proto.state = MagicMock()
        self.addCleanup(self.proto.factory.sessionhandler.disconnect_all)
        DelayedCall.debug = True

    def tearDown(self):
        super().tearDown()

    @mock.patch("evennia.server.portal.portalsessionhandler.reactor", new=MagicMock())
    def test_data_in(self):
        self.proto.sessionhandler.data_in = MagicMock()
        self.proto.onOpen()
        msg = json.dumps(["logged_in", (), {}]).encode()
        self.proto.onMessage(msg, isBinary=False)
        self.proto.sessionhandler.data_in.assert_called_with(self.proto, logged_in=[[], {}])
        sendStr = "You can get anything you want at Alice's Restaurant."
        msg = json.dumps(["text", (sendStr,), {}]).encode()
        self.proto.onMessage(msg, isBinary=False)
        self.proto.sessionhandler.data_in.assert_called_with(self.proto, text=[[sendStr], {}])

    @mock.patch("evennia.server.portal.portalsessionhandler.reactor", new=MagicMock())
    def test_data_out(self):
        self.proto.onOpen()
        self.proto.sendLine = MagicMock()
        msg = json.dumps(["logged_in", (), {}])
        self.proto.sessionhandler.data_out(self.proto, text=[["Excepting Alice"], {}])
        self.proto.sendLine.assert_called_with(json.dumps(["text", ["Excepting Alice"], {}]))



"""
Red Button

Griatch - 2011

This is a more advanced example object. It combines functions from
script.examples as well as commands.examples to make an interactive
button typeclass.

Create this button with

    create/drop tutorials.red_button.RedButton

Note that you must drop the button before you can see its messages!

## Technical

The button's functionality is controlled by CmdSets that gets added and removed
depending on the 'state' the button is in.

- Lid-closed state: In this state the button is covered by a glass cover and trying
  to 'push' it will fail. You can 'nudge', 'smash' or 'open' the lid.
- Lid-open state: In this state the lid is open but will close again after a certain
  time. Using 'push' now will press the button and trigger the Blind-state.
- Blind-state: In this mode you are blinded by a bright flash. This will affect your
  normal commands like 'look' and help until the blindness wears off after a certain
  time.

Timers are handled by persistent delays on the button. These are examples of
`evennia.utils.utils.delay` calls that wait a certain time before calling a method -
such as when closing the lid and un-blinding a character.

"""
import random

from evennia import CmdSet, Command, DefaultObject
from evennia.utils.utils import delay, interactive, repeat

# Commands on the button (not all awailable at the same time)


# Commands for the state when the lid covering the button is closed.


class CmdPushLidClosed(Command):
    """
    Push the red button (lid closed)

    Usage:
      push button

    """

    key = "push button"
    aliases = ["push", "press button", "press"]
    locks = "cmd:all()"

    def func(self):
        """
        This is the version of push used when the lid is closed.

        An alternative would be to make a 'push' command in a default cmdset
        that is always available on the button and then use if-statements to
        check if the lid is open or closed.

        """
        self.caller.msg("You cannot push the button = there is a glass lid covering it.")


class CmdNudge(Command):
    """
    Try to nudge the button's lid.

    Usage:
      nudge lid

    This command will have you try to push the lid of the button away.

    """

    key = "nudge lid"  # two-word command name!
    aliases = ["nudge"]
    locks = "cmd:all()"

    def func(self):
        """
        Nudge the lid. Random chance of success to open it.

        """
        rand = random.random()
        if rand < 0.5:
            self.caller.msg("You nudge at the lid. It seems stuck.")
        elif rand < 0.7:
            self.caller.msg("You move the lid back and forth. It won't budge.")
        else:
            self.caller.msg("You manage to get a nail under the lid.")
            # self.obj is the button object
            self.obj.to_open_state()


class CmdSmashGlass(Command):
    """
    Smash the protective glass.

    Usage:
      smash glass

    Try to smash the glass of the button.

    """

    key = "smash glass"
    aliases = ["smash lid", "break lid", "smash"]
    locks = "cmd:all()"

    def func(self):
        """
        The lid won't open, but there is a small chance of causing the lamp to
        break.

        """
        rand = random.random()
        self.caller.location.msg_contents(
            f"{self.caller.name} tries to smash the glass of the button.", exclude=self.caller
        )

        if rand < 0.2:
            string = (
                "You smash your hand against the glass"
                " with all your might. The lid won't budge"
                " but you cause quite the tremor through the button's mount."
                "\nIt looks like the button's lamp stopped working for the time being, "
                "but the lid is still as closed as ever."
            )
            # self.obj is the button itself
            self.obj.break_lamp()
        elif rand < 0.6:
            string = "You hit the lid hard. It doesn't move an inch."
        else:
            string = (
                "You place a well-aimed fist against the glass of the lid."
                " Unfortunately all you get is a pain in your hand. Maybe"
                " you should just try to just ... open the lid instead?"
            )
        self.caller.msg(string)


class CmdOpenLid(Command):
    """
    open lid

    Usage:
      open lid

    """

    key = "open lid"
    aliases = ["open button"]
    locks = "cmd:all()"

    def func(self):
        "simply call the right function."

        if self.obj.db.lid_locked:
            self.caller.msg("This lid seems locked in place for the moment.")
            return

        string = "\nA ticking sound is heard, like a winding mechanism. Seems "
        string += "the lid will soon close again."
        self.caller.msg(string)
        self.caller.location.msg_contents(
            f"{self.caller.name} opens the lid of the button.", exclude=self.caller
        )
        self.obj.to_open_state()


class LidClosedCmdSet(CmdSet):
    """
    A simple cmdset tied to the redbutton object.

    It contains the commands that launches the other
    command sets, making the red button a self-contained
    item (i.e. you don't have to manually add any
    scripts etc to it when creating it).

    Note that this is given with a `key_mergetype` set. This
    is set up so that the cmdset with merge with Union merge type
    *except* if the other cmdset to merge with is LidOpenCmdSet,
    in which case it will Replace that. So these two cmdsets will
    be mutually exclusive.

    """

    key = "LidClosedCmdSet"

    def at_cmdset_creation(self):
        "Populates the cmdset when it is instantiated."
        self.add(CmdPushLidClosed())
        self.add(CmdNudge())
        self.add(CmdSmashGlass())
        self.add(CmdOpenLid())


# Commands for the state when the button's protective cover is open - now the
# push command will work. You can also close the lid again.


class CmdPushLidOpen(Command):
    """
    Push the red button

    Usage:
      push button

    """

    key = "push button"
    aliases = ["push", "press button", "press"]
    locks = "cmd:all()"

    @interactive
    def func(self):
        """
        This version of push will immediately trigger the next button state.

        The use of the @interactive decorator allows for using `yield` to add
        simple pauses in how quickly a message is returned to the user.  This
        kind of pause will not survive a server reload.

        """
        # pause a little between each message.
        self.caller.msg("You reach out to press the big red button ...")
        yield (2)  # pause 2s before next message
        self.caller.msg("\n\n|wBOOOOM! A bright light blinds you!|n")
        yield (1)  # pause 1s before next message
        self.caller.msg("\n\n|xThe world goes dark ...|n")

        name = self.caller.name
        self.caller.location.msg_contents(
            f"{name} presses the button. BOOM! {name} is blinded by a flash!", exclude=self.caller
        )
        self.obj.blind_target(self.caller)


class CmdCloseLid(Command):
    """
    Close the lid

    Usage:
      close lid

    Closes the lid of the red button.
    """

    key = "close lid"
    aliases = ["close"]
    locks = "cmd:all()"

    def func(self):
        "Close the lid"

        self.obj.to_closed_state()

        # this will clean out scripts dependent on lid being open.
        self.caller.msg("You close the button's lid. It clicks back into place.")
        self.caller.location.msg_contents(
            f"{self.caller.name} closes the button's lid.", exclude=self.caller
        )


class LidOpenCmdSet(CmdSet):
    """
    This is the opposite of the Closed cmdset.

    Note that this is given with a `key_mergetype` set. This
    is set up so that the cmdset with merge with Union merge type
    *except* if the other cmdset to merge with is LidClosedCmdSet,
    in which case it will Replace that. So these two cmdsets will
    be mutually exclusive.

    """

    key = "LidOpenCmdSet"

    def at_cmdset_creation(self):
        """Setup the cmdset"""
        self.add(CmdPushLidOpen())
        self.add(CmdCloseLid())


# Commands for when the button has been pushed and the player is blinded. This
# replaces commands on the player making them 'blind' for a while.


class CmdBlindLook(Command):
    """
    Looking around in darkness

    Usage:
      look <obj>

    ... not that there's much to see in the dark.

    """

    key = "look"
    aliases = ["l", "get", "examine", "ex", "feel", "listen"]
    locks = "cmd:all()"

    def func(self):
        "This replaces all the senses when blinded."

        # we decide what to reply based on which command was
        # actually tried

        if self.cmdstring == "get":
            string = "You fumble around blindly without finding anything."
        elif self.cmdstring == "examine":
            string = "You try to examine your surroundings, but can't see a thing."
        elif self.cmdstring == "listen":
            string = "You are deafened by the boom."
        elif self.cmdstring == "feel":
            string = "You fumble around, hands outstretched. You bump your knee."
        else:
            # trying to look
            string = (
                "You are temporarily blinded by the flash. "
                "Until it wears off, all you can do is feel around blindly."
            )
        self.caller.msg(string)
        self.caller.location.msg_contents(
            f"{self.caller.name} stumbles around, blinded.", exclude=self.caller
        )


class CmdBlindHelp(Command):
    """
    Help function while in the blinded state

    Usage:
      help

    """

    key = "help"
    aliases = "h"
    locks = "cmd:all()"

    def func(self):
        """
        Just give a message while blinded. We could have added this to the
        CmdBlindLook command too if we wanted to keep things more compact.

        """
        self.caller.msg("You are beyond help ... until you can see again.")


class BlindCmdSet(CmdSet):
    """
    This is the cmdset added to the *account* when
    the button is pushed.

    Since this has mergetype Replace it will completely remove the commands of
    all other cmdsets while active. To allow some limited interaction
    (pose/say) we import those default commands and add them too.

    We also disable all exit-commands generated by exits and
    object-interactions while blinded by setting `no_exits` and `no_objs` flags
    on the cmdset. This is to avoid the player walking off or interfering with
    other objects while blinded. Account-level commands however (channel messaging
    etc) will not be affected by the blinding.

    """

    key = "BlindCmdSet"
    # we want it to completely replace all normal commands
    # until the timed script removes it again.
    mergetype = "Replace"
    # we want to stop the player from walking around
    # in this blinded state, so we hide all exits too.
    # (channel commands will still work).
    no_exits = True  # keep player in the same room
    no_objs = True  # don't allow object commands

    def at_cmdset_creation(self):
        "Setup the blind cmdset"
        from evennia.commands.default.general import CmdPose, CmdSay

        self.add(CmdSay())
        self.add(CmdPose())
        self.add(CmdBlindLook())
        self.add(CmdBlindHelp())


#
# Definition of the object itself
#


class RedButton(DefaultObject):
    """
    This class describes an evil red button.  It will blink invitingly and
    temporarily blind whomever presses it.

    The button can take a few optional attributes controlling how things will
    be displayed in its various states. This is a useful way to give builders
    the option to customize a complex object from in-game. Actual return messages
    to event-actions are (in this example) left with each command, but one could
    also imagine having those handled via Attributes as well, if one wanted a
    completely in-game customizable button without needing to tweak command
    classes.

    Attributes:
    - `desc_closed_lid`:  This is the description to show of the button
      when the lid is closed.
    - `desc_open_lid`": Shown when the lid is open
    - `auto_close_msg`: Message to show when lid auto-closes
    - `desc_add_lamp_broken`: Extra desc-line added after normal desc when lamp
      is broken.
    - blink_msg: A list of strings to randomly choose from when the lamp
      blinks.

    Notes:
    The button starts with lid closed. To set the initial description,
    you can either set desc after creating it or pass a `desc` attribute
    when creating it, such as
    `button = create_object(RedButton, ..., attributes=[('desc', 'my desc')])`.

    """

    # these are the pre-set descriptions. Setting attributes will override
    # these on the fly.

    desc_closed_lid = (
        "This is a large red button, inviting yet evil-looking. A closed glass lid protects it."
    )
    desc_open_lid = (
        "This is a large red button, inviting yet evil-looking. "
        "Its glass cover is open and the button exposed."
    )
    auto_close_msg = "The button's glass lid silently slides back in place."
    lamp_breaks_msg = "The lamp flickers, the button going dark."
    desc_add_lamp_broken = "\nThe big red button has stopped blinking for the time being."
    # note that this is a list. A random message will display each time
    blink_msgs = [
        "The red button flashes briefly.",
        "The red button blinks invitingly.",
        "The red button flashes. You know you wanna push it!",
    ]

    def at_object_creation(self):
        """
        This function is called (once) when object is created.

        """
        self.db.lamp_works = True

        # start closed
        self.to_closed_state()

        # start blinking every 35s.
        repeat(35, self._do_blink, persistent=True)

    def _do_blink(self):
        """
        Have the button blink invitingly unless it's broken.

        """
        if self.location and self.db.lamp_works:
            possible_messages = self.db.blink_msgs or self.blink_msgs
            self.location.msg_contents(random.choice(possible_messages))

    def _set_desc(self, attrname=None):
        """
        Set a description, based on the attrname given, taking the lamp-status
        into account.

        Args:
            attrname (str, optional): This will first check for an Attribute with this name,
                secondly for a property on the class. So if `attrname="auto_close_msg"`,
                we will first look for an attribute `.db.auto_close_msg` and if that's
                not found we'll use `.auto_close_msg` instead. If unset (`None`), the
                currently set desc will not be changed (only lamp will be checked).

        Notes:
            If `self.db.lamp_works` is `False`, we'll append
            `desc_add_lamp_broken` text.

        """
        if attrname:
            # change desc
            desc = self.attributes.get(attrname) or getattr(self, attrname)
        else:
            # use existing desc
            desc = self.db.desc

        if not self.db.lamp_works:
            # lamp not working. Add extra to button's desc
            desc += self.db.desc_add_lamp_broken or self.desc_add_lamp_broken

        self.db.desc = desc

    # state-changing methods and actions

    def to_closed_state(self, msg=None):
        """
        Switches the button to having its lid closed.

        Args:
            msg (str, optional): If given, display a message to the room
            when lid closes.

        This will first try to get the Attribute (self.db.desc_closed_lid) in
        case it was set by a builder and if that was None, it will fall back to
        self.desc_closed_lid, the default description (note that lack of .db).
        """
        self._set_desc("desc_closed_lid")
        # remove lidopen-state, if it exists
        self.cmdset.remove(LidOpenCmdSet)
        # add lid-closed cmdset
        self.cmdset.add(LidClosedCmdSet, persistent=True)

        if msg and self.location:
            self.location.msg_contents(msg)

    def to_open_state(self):
        """
        Switches the button to having its lid open. This also starts a timer
        that will eventually close it again.

        """
        self._set_desc("desc_open_lid")
        # remove lidopen-state, if it exists
        self.cmdset.remove(LidClosedCmdSet)
        # add lid-open cmdset
        self.cmdset.add(LidOpenCmdSet, persistent=True)

        # wait 20s then call self.to_closed_state with a message as argument
        delay(
            35, self.to_closed_state, self.db.auto_close_msg or self.auto_close_msg, persistent=True
        )

    def _unblind_target(self, caller):
        """
        This is called to un-blind after a certain time.

        """
        caller.cmdset.remove(BlindCmdSet)
        caller.msg("You blink feverishly as your eyesight slowly returns.")
        self.location.msg_contents(
            f"{caller.name} seems to be recovering their eyesight, blinking feverishly.",
            exclude=caller,
        )

    def blind_target(self, caller):
        """
        Someone was foolish enough to press the button! Blind them
        temporarily.

        Args:
            caller (Object): The one to be blinded.

        """

        # we don't need to remove other cmdsets, this will replace all,
        # then restore whatever was there when it goes away.
        # we don't make this persistent, to make sure any problem is just a reload away
        caller.cmdset.add(BlindCmdSet)

        # wait 20s then call self._unblind to remove blindness effect. The
        # persistent=True means the delay should survive a server reload.
        delay(20, self._unblind_target, caller, persistent=True)

    def _unbreak_lamp(self):
        """
        This is called to un-break the lamp after a certain time.

        """
        # we do this quietly, the user will just notice it starting blinking again
        self.db.lamp_works = True
        self._set_desc()

    def break_lamp(self):
        """
        Breaks the lamp in the button, stopping it from blinking for a while

        """
        self.db.lamp_works = False
        # this will update the desc with the info about the broken lamp
        self._set_desc()
        self.location.msg_contents(self.db.lamp_breaks_msg or self.lamp_breaks_msg)

        # wait 21s before unbreaking the lamp again
        delay(21, self._unbreak_lamp)



"""
Views for managing channels.

"""

from django.conf import settings
from django.db.models.functions import Lower
from django.http import HttpResponseBadRequest
from django.utils.text import slugify
from django.views.generic import ListView

from evennia.utils import class_from_module
from evennia.utils.logger import tail_log_file

from .mixins import TypeclassMixin
from .objects import ObjectDetailView


class ChannelMixin(TypeclassMixin):
    """
    This is a "mixin", a modifier of sorts.

    Any view class with this in its inheritance list will be modified to work
    with HelpEntry objects instead of generic Objects or otherwise.

    """

    # -- Django constructs --
    model = class_from_module(
        settings.BASE_CHANNEL_TYPECLASS, fallback=settings.FALLBACK_CHANNEL_TYPECLASS
    )

    # -- Evennia constructs --
    page_title = "Channels"

    # What lock type to check for the requesting user, authenticated or not.
    # https://github.com/evennia/evennia/wiki/Locks#valid-access_types
    access_type = "listen"

    def get_queryset(self):
        """
        Django hook; here we want to return a list of only those Channels
        and other documentation that the current user is allowed to see.

        Returns:
            queryset (QuerySet): List of Channels available to the user.

        """
        account = self.request.user

        # Get list of all Channels
        channels = self.typeclass.objects.all().iterator()

        # Now figure out which ones the current user is allowed to see
        bucket = [channel.id for channel in channels if channel.access(account, "listen")]

        # Re-query and set a sorted list
        filtered = self.typeclass.objects.filter(id__in=bucket).order_by(Lower("db_key"))

        return filtered


class ChannelListView(ChannelMixin, ListView):
    """
    Returns a list of channels that can be viewed by a user, authenticated
    or not.

    """

    # -- Django constructs --
    paginate_by = 100
    template_name = "website/channel_list.html"

    # -- Evennia constructs --
    page_title = "Channel Index"

    max_popular = 10

    def get_context_data(self, **kwargs):
        """
        Django hook; we override it to calculate the most popular channels.

        Returns:
            context (dict): Django context object

        """
        context = super().get_context_data(**kwargs)

        # Calculate which channels are most popular
        context["most_popular"] = sorted(
            list(self.get_queryset()),
            key=lambda channel: len(channel.subscriptions.all()),
            reverse=True,
        )[: self.max_popular]

        return context


class ChannelDetailView(ChannelMixin, ObjectDetailView):
    """
    Returns the log entries for a given channel.

    """

    # -- Django constructs --
    template_name = "website/channel_detail.html"

    # -- Evennia constructs --
    # What attributes of the object you wish to display on the page. Model-level
    # attributes will take precedence over identically-named db.attributes!
    # The order you specify here will be followed.
    attributes = ["name"]

    # How many log entries to read and display.
    max_num_lines = 10000

    def get_context_data(self, **kwargs):
        """
        Django hook; before we can display the channel logs, we need to recall
        the logfile and read its lines.

        Returns:
            context (dict): Django context object

        """
        # Get the parent context object, necessary first step
        context = super().get_context_data(**kwargs)
        channel = self.object

        # Get the filename this Channel is recording to
        filename = channel.get_log_filename()

        # Split log entries so we can filter by time
        bucket = []
        for log in (x.strip() for x in tail_log_file(filename, 0, self.max_num_lines)):
            if not log:
                continue
            try:
                time, msg = log.split(" [-] ")
                time_key = time.split(":")[0]
            except ValueError:
                # malformed log line. Skip.
                continue

            bucket.append({"key": time_key, "timestamp": time, "message": msg})

        # Add the processed entries to the context
        context["object_list"] = bucket

        # Get a list of unique timestamps by hour and sort them
        context["object_filters"] = sorted(set([x["key"] for x in bucket]))

        return context

    def get_object(self, queryset=None):
        """
        Override of Django hook that retrieves an object by slugified channel
        name.

        Returns:
            channel (Channel): Channel requested in the URL.

        """
        # Get the queryset for the help entries the user can access
        if not queryset:
            queryset = self.get_queryset()

        # Find the object in the queryset
        channel = slugify(self.kwargs.get("slug", ""))
        obj = next((x for x in queryset if slugify(x.db_key) == channel), None)

        # Check if this object was requested in a valid manner
        if not obj:
            raise HttpResponseBadRequest(
                "No %(verbose_name)s found matching the query"
                % {"verbose_name": queryset.model._meta.verbose_name}
            )

        return obj



"""
TTYPE (MTTS) - Mud Terminal Type Standard

This module implements the TTYPE telnet protocol as per
http://tintin.sourceforge.net/mtts/. It allows the server to ask the
client about its capabilities. If the client also supports TTYPE, it
will return with information such as its name, if it supports colour
etc. If the client does not support TTYPE, this will be ignored.

All data will be stored on the protocol's protocol_flags dictionary,
under the 'TTYPE' key.

"""

# telnet option codes
TTYPE = bytes([24])  # b"\x18"
IS = bytes([0])  # b"\x00"
SEND = bytes([1])  # b"\x01"

# terminal capabilities and their codes
MTTS = [
    (2048, "SSL"),
    (1024, "MSLP"),
    (512, "MNES"),
    (256, "TRUECOLOR"),
    (128, "PROXY"),
    (64, "SCREENREADER"),
    (32, "OSC_COLOR_PALETTE"),
    (16, "MOUSE_TRACKING"),
    (8, "XTERM256"),
    (4, "UTF-8"),
    (2, "VT100"),
    (1, "ANSI"),
]


class Ttype:
    """
    Handles ttype negotiations. Called and initiated by the
    telnet protocol.

    """

    def __init__(self, protocol):
        """
        Initialize ttype by storing protocol on ourselves and calling
        the client to see if it supporst ttype.

        Args:
            protocol (Protocol): The protocol instance.

        Notes:
            The `self.ttype_step` indicates how far in the data
            retrieval we've gotten.

        """
        self.ttype_step = 0
        self.protocol = protocol
        # we set FORCEDENDLINE for clients not supporting ttype
        self.protocol.protocol_flags["FORCEDENDLINE"] = True
        self.protocol.protocol_flags["TTYPE"] = False
        # is it a safe bet to assume ANSI is always supported?
        self.protocol.protocol_flags["ANSI"] = True
        # setup protocol to handle ttype initialization and negotiation
        self.protocol.negotiationMap[TTYPE] = self.will_ttype
        # ask if client will ttype, connect callback if it does.
        self.protocol.do(TTYPE).addCallbacks(self.will_ttype, self.wont_ttype)

    def wont_ttype(self, option):
        """
        Callback if ttype is not supported by client.

        Args:
            option (Option): Not used.

        """
        self.protocol.protocol_flags["TTYPE"] = False
        self.protocol.handshake_done()

    def will_ttype(self, option):
        """
        Handles negotiation of the ttype protocol once the client has
        confirmed that it will respond with the ttype protocol.

        Args:
            option (Option): Not used.

        Notes:
            The negotiation proceeds in several steps, each returning a
            certain piece of information about the client. All data is
            stored on protocol.protocol_flags under the TTYPE key.

        """
        options = self.protocol.protocol_flags

        if options and options.get("TTYPE", False) or self.ttype_step > 3:
            return

        try:
            option = b"".join(option).lstrip(IS).decode()
        except TypeError:
            # option is not on a suitable form for joining
            pass

        if self.ttype_step == 0:
            # just start the request chain
            self.protocol.requestNegotiation(TTYPE, SEND)

        elif self.ttype_step == 1:
            # this is supposed to be the name of the client/terminal.
            # For clients not supporting the extended TTYPE
            # definition, subsequent calls will just repeat-return this.
            try:
                clientname = option.upper()
            except AttributeError:
                # malformed option (not a string)
                clientname = "UNKNOWN"

            # use name to identify support for xterm256. Many of these
            # only support after a certain version, but all support
            # it since at least 4 years. We assume recent client here for now.
            xterm256 = False
            if clientname.startswith("MUDLET"):
                # supports xterm256 stably since 1.1 (2010?)
                xterm256 = clientname.split("MUDLET", 1)[1].strip() >= "1.1"
                # Mudlet likes GA's on a prompt line for the prompt trigger to
                # match, if it's not wanting NOGOAHEAD.
                if not self.protocol.protocol_flags["NOGOAHEAD"]:
                    self.protocol.protocol_flags["NOGOAHEAD"] = True
                    self.protocol.protocol_flags["NOPROMPTGOAHEAD"] = False

            if (
                clientname.startswith("XTERM")
                or clientname.endswith("-256COLOR")
                or clientname
                in (
                    "ATLANTIS",  # > 0.9.9.0 (aug 2009)
                    "CMUD",  # > 3.04 (mar 2009)
                    "KILDCLIENT",  # > 2.2.0 (sep 2005)
                    "MUDLET",  # > beta 15 (sep 2009)
                    "MUSHCLIENT",  # > 4.02 (apr 2007)
                    "PUTTY",  # > 0.58 (apr 2005)
                    "BEIP",  # > 2.00.206 (late 2009) (BeipMu)
                    "POTATO",  # > 2.00 (maybe earlier)
                    "TINYFUGUE",  # > 4.x (maybe earlier)
                )
            ):
                xterm256 = True

            # all clients supporting TTYPE at all seem to support ANSI
            self.protocol.protocol_flags["ANSI"] = True
            self.protocol.protocol_flags["XTERM256"] = xterm256
            self.protocol.protocol_flags["CLIENTNAME"] = clientname
            self.protocol.requestNegotiation(TTYPE, SEND)

        elif self.ttype_step == 2:
            # this is a term capabilities flag
            term = option
            tupper = term.upper()
            # identify xterm256 based on flag
            xterm256 = (
                tupper.endswith("-256COLOR")
                or tupper.endswith("XTERM")  # Apple Terminal, old Tintin
                and not tupper.endswith("-COLOR")  # old Tintin, Putty
            )
            if xterm256:
                self.protocol.protocol_flags["ANSI"] = True
                self.protocol.protocol_flags["XTERM256"] = xterm256
            self.protocol.protocol_flags["TERM"] = term
            # request next information
            self.protocol.requestNegotiation(TTYPE, SEND)

        elif self.ttype_step == 3:
            # the MTTS bitstring identifying term capabilities
            if option.startswith("MTTS"):
                option = option[4:].strip()
                if option.isdigit():
                    # a number - determine the actual capabilities
                    option = int(option)
                    support = dict(
                        (capability, True) for bitval, capability in MTTS if option & bitval > 0
                    )
                    self.protocol.protocol_flags.update(support)
                else:
                    # some clients send erroneous MTTS as a string. Add directly.
                    self.protocol.protocol_flags[option.upper()] = True

            self.protocol.protocol_flags["TTYPE"] = True
            # we must sync ttype once it'd done
            self.protocol.handshake_done()
        self.ttype_step += 1



"""
NPC AI module for EvAdventure (WIP)

This implements a state machine for the NPCs, where it uses inputs from the game to determine what
to do next. The AI works on the concept of being 'ticks', at which point, the AI will decide to move
between different 'states', performing different 'actions' within each state until changing to
another state. The odds of changing between states and performing actions are weighted, allowing for
an AI agent to be more or less likely to perform certain actions.

The state machine is fed a dictionary of states and their transitions, and a dictionary of available
actions to choose between.
::

    {
        "states": {
            "state1": {"action1": odds, "action2": odds, ...},
            "state2": {"action1": odds, "action2": odds, ...}, ...
        }
        "transition": {
            "state1": {"state2": "odds, "state3": odds, ...},
            "state2": {"state1": "odds, "state3": odds, ...}, ...
        }
    }

The NPC class needs to look like this:
::

    class NPC(DefaultCharacter):

        # ...

        @lazy_property
        def ai(self):
            return AIHandler(self)

        def ai_roam(self, action):
            # perform the action within the current state ai.state

        def ai_hunt(self, action):
            # etc

"""

import random

from evennia.utils import logger
from evennia.utils.dbserialize import deserialize

# Some example AI structures

EMOTIONAL_AI = {
    # Non-combat AI that has different moods for conversations
    "states": {
        "neutral": {"talk_neutral": 0.9, "change_state": 0.1},
        "happy": {"talk_happy": 0.9, "change_state": 0.1},
        "sad": {"talk_sad": 0.9, "change_state": 0.1},
        "angry": {"talk_angry": 0.9, "change_state": 0.1},
    }
}

STATIC_AI = {
    # AI that just hangs around until attacked
    "states": {
        "idle": {"do_nothing": 1.0},
        "combat": {"attack": 0.9, "stunt": 0.1},
    }
}

ROAM_AI = {
    # AI that roams around randomly, now and then stopping.
    "states": {
        "idle": {"do_nothing": 0.9, "change_state": 0.1},
        "roam": {
            "move_north": 0.1,
            "move_south": 0.1,
            "move_east": 0.1,
            "move_west": 0.1,
            "wait": 0.4,
            "change_state": 0.2,
        },
        "combat": {"attack": 0.9, "stunt": 0.05, "flee": 0.05},
    },
    "transitions": {
        "idle": {"roam": 0.5, "idle": 0.5},
        "roam": {"idle": 0.1, "roam": 0.9},
    },
}

HUNTER_AI = {
    "states": {
        "hunt_roam": {
            "move_north": 0.2,
            "move_south": 0.2,
            "move_east": 0.2,
            "move_west": 0.2,
        },
        "hunt_track": {
            "track_and_move": 0.9,
            "change_state": 0.1,
        },
        "combat": {"attack": 0.8, "stunt": 0.1, "other": 0.1},
    },
    "transitions": {
        # add a chance of the hunter losing its trail
        "hunt_track": {"hunt_roam": 1.0},
    },
}


class AIHandler:
    """
    AIHandler class. This should be placed on the NPC object, and will handle the state machine,
    including transitions and actions.

    Add to typeclass with @lazyproperty:

        class NPC(DefaultCharacter):

            ai_states = {...}

            # ...

            @lazyproperty
            def ai(self):
                return AIHandler(self)

    """

    def __init__(self, obj):
        self.obj = obj

        if hasattr(self, "ai_states"):
            # since we're not setting `force=True` here, we won't overwrite any existing /
            # customized dicts.
            self.add_aidict(self.ai_states)

    def __str__(self):
        return f"AIHandler for {self.obj}. Current state: {self.state}"

    @staticmethod
    def _normalize_odds(odds):
        """
        Normalize odds to 1.0.

        Args:
            odds (list): List of odds to normalize.
        Returns:
            list: Normalized list of odds.

        """
        return [float(i) / sum(odds) for i in odds]

    @staticmethod
    def _weighted_choice(choices, odds):
        """
        Choose a random element from a list of choices, with odds.

        Args:
            choices (list): List of choices to choose from. Unordered.
            odds (list): List of odds to choose from, matching the choices list. This
                can be a list of integers or floats, indicating priority. Have odds sum
                up to 100 or 1.0 to properly represent predictable odds.
        Returns:
            object: Randomly chosen element from choices.

        """
        if choices:
            return random.choices(choices, odds)[0]

    @staticmethod
    def _weighted_choice_dict(choices):
        """
        Choose a random element from a dictionary of choices, with odds.

        Args:
            choices (dict): Dictionary of choices to choose from, with odds as values.
        Returns:
            object: Randomly chosen element from choices.

        """
        return AIHandler._weighted_choice(list(choices.keys()), list(choices.values()))

    @staticmethod
    def _validate_ai_dict(aidict):
        """
        Validate and normalize an AI dictionary.

        Args:
            aidict (dict): AI dictionary to normalize.
        Returns:
            dict: Normalized AI dictionary.

        """
        if "states" not in aidict:
            raise ValueError("AI dictionary must contain a 'states' key.")

        if "transitions" not in aidict:
            aidict["transitions"] = {}

        # if we have no transitions, make sure we have a transition for each state set to 0
        for state in aidict["states"]:
            if state not in aidict["transitions"]:
                aidict["transitions"][state] = {}
            for state2 in aidict["states"]:
                if state2 not in aidict["transitions"][state]:
                    aidict["transitions"][state][state2] = 0.0

        # normalize odds
        for state, actions in aidict["states"].items():
            aidict["states"][state] = AIHandler._normalize_odds(list(actions.values()))
        for state, transitions in aidict["transitions"].items():
            aidict["transitions"][state] = AIHandler._normalize_odds(list(transitions.values()))

        return aidict

    @property
    def state(self):
        """
        Return the current state of the AI.

        Returns:
            str: Current state of the AI.

        """
        return self.obj.attributes.get("ai_state", category="ai", default="idle")

    @state.setter
    def state(self, value):
        """
        Set the current state of the AI. This allows to force a state change, e.g. when starting
        combat.

        Args:
            value (str): New state of the AI.

        """
        return self.obj.attributes.add("ai_state", category="ai")

    @property
    def states(self):
        """
        Return the states dictionary for the AI.

        Returns:
            dict: States dictionary for the AI.

        """
        return self.obj.attributes.get("ai_states", category="ai", default={"idle": {}})

    @states.setter
    def states(self, value):
        """
        Set the states dictionary for the AI.

        Args:
            value (dict): New states dictionary for the AI.

        """
        return self.obj.attributes.add("ai_states", value, category="ai")

    @property
    def transitions(self):
        """
        Return the transitions dictionary for the AI.

        Returns:
            dict: Transitions dictionary for the AI.

        """
        return self.obj.attributes.get("ai_transitions", category="ai", default={"idle": []})

    @transitions.setter
    def transitions(self, value):
        """
        Set the transitions dictionary for the AI.

        Args:
            value (dict): New transitions dictionary for the AI. This will be automatically
            normalized.

        """
        for state in value.keys():
            value[state] = dict(
                zip(value[state].keys(), self._normalize_odds(value[state].values()))
            )
        return self.obj.attributes.add("ai_transitions", value, category="ai")

    def add_aidict(self, aidict, force=False):
        """
        Add an AI dictionary to the AI handler, if one doesn't already exist.

        Args:
            aidict (dict): AI dictionary to add.
            force (bool, optional): Force adding the AI dictionary, even if one already exists on
            this handler.

        """
        if not force and self.states and self.transitions:
            return

        aidict = self._validate_ai_dict(aidict)
        self.states = aidict["states"]
        self.transitions = aidict["transitions"]

    def adjust_transition_probability(self, state_start, state_end, odds):
        """
        Adjust the transition probability between two states.

        Args:
            state_start (str): State to start from.
            state_end (str): State to end at.
            odds (int): New odds for the transition.

        Note:
            This will normalize the odds across the other transitions from the starting state.

        """
        transitions = deserialize(self.transitions)
        transitions[state_start][state_end] = odds
        transitions[state_start] = dict(
            zip(
                transitions[state_start].keys(),
                self._normalize_odds(transitions[state_start].values()),
            )
        )
        self.transitions = transitions

    def get_next_state(self):
        """
        Get the next state for the AI.

        Returns:
            str: Next state for the AI.

        """
        return self._weighted_choice_dict(self.transitions[self.state])

    def get_next_action(self):
        """
        Get the next action for the AI within the current state.

        Returns:
            str: Next action for the AI.

        """
        return self._weighted_choice_dict(self.states[self.state])

    def execute_ai(self):
        """
        Execute the next ai action in the current state.

        This assumes that each available state exists as a method on the object, named
        ai_<state_name>, taking an optional argument of the next action to perform. The method
        will itself update the state or transition weights through this handler.

        Some states have in-built state transitions, via the special "change_state" action.

        """
        next_action = self.get_next_action()
        statechange = 0
        while next_action == "change_state":
            self.state = self.get_next_state()
            next_action = self.get_next_action()
            if statechange > 5:
                logger.log_err(f"AIHandler: {self.obj} got stuck in a state-change loop.")
                return

        # perform the action
        try:
            getattr(self.obj, f"ai_{self.state}")(next_action)
        except AttributeError:
            logger.log_err(f"AIHandler: {self.obj} has no ai_{self.state} method.")



"""
Intro menu / game tutor

Evennia contrib - Griatch 2020

This contrib is an intro-menu for general MUD and evennia usage using the
EvMenu menu-templating system.

EvMenu templating is a way to create a menu using a string-format instead
of creating all nodes manually. Of course, for full functionality one must
still create the goto-callbacks.

"""

from evennia import CmdSet, create_object
from evennia.utils.evmenu import EvMenu, parse_menu_template

# Goto callbacks and helper resources for the menu


def do_nothing(caller, raw_string, **kwargs):
    """
    Re-runs the current node
    """
    return None


def send_testing_tagged(caller, raw_string, **kwargs):
    """
    Test to send a message to a pane tagged with 'testing' in the webclient.

    """
    caller.msg(
        (
            (
                "This is a message tagged with 'testing' and "
                "should appear in the pane you selected!\n "
                f"You wrote: '{raw_string}'"
            ),
            {"type": "testing"},
        )
    )
    return None


# Resources for the first help-command demo


class DemoCommandSetHelp(CmdSet):
    """
    Demo the help command
    """

    key = "Help Demo Set"
    priority = 2

    def at_cmdset_creation(self):
        from evennia import default_cmds

        self.add(default_cmds.CmdHelp())
        self.add(default_cmds.CmdChannel())


def goto_command_demo_help(caller, raw_string, **kwargs):
    "Sets things up before going to the help-demo node"
    _maintain_demo_room(caller, delete=True)
    caller.cmdset.remove(DemoCommandSetRoom)
    caller.cmdset.remove(DemoCommandSetComms)
    caller.cmdset.add(DemoCommandSetHelp)  # TODO - make persistent
    return kwargs.get("gotonode") or "command_demo_help"


# Resources for the comms demo


class DemoCommandSetComms(CmdSet):
    """
    Demo communications
    """

    key = "Color Demo Set"
    priority = 2
    no_exits = True
    no_objs = True

    def at_cmdset_creation(self):
        from evennia import default_cmds

        self.add(default_cmds.CmdHelp())
        self.add(default_cmds.CmdSay())
        self.add(default_cmds.CmdPose())
        self.add(default_cmds.CmdPage())
        self.add(default_cmds.CmdColorTest())


def goto_command_demo_comms(caller, raw_string, **kwargs):
    """
    Setup and go to the color demo node.
    """
    caller.cmdset.remove(DemoCommandSetHelp)
    caller.cmdset.remove(DemoCommandSetRoom)
    caller.cmdset.add(DemoCommandSetComms)
    return kwargs.get("gotonode") or "comms_demo_start"


# Resources for the room demo

_ROOM_DESC = """
This is a small and comfortable wood cabin. Bright sunlight is shining in
through the windows.

Use |ylook sign|n or |yl sign|n to examine the wooden sign nailed to the wall.

"""

_SIGN_DESC = """
The small sign reads:

    Good! Now try '|ylook small|n'.

    ... You'll get a multi-match error! There are two things that 'small' could
    refer to here - the 'small wooden sign' or the 'small, cozy cabin' itself.  You will
    get a list of the possibilities.

    You could either tell Evennia which one you wanted by picking a unique part
    of their name (like '|ylook cozy|n') or use the number in the list to pick
    the one you want, like this:

        |ylook small-2|n

    As long as what you write is uniquely identifying you can be lazy and not
    write the full name of the thing you want to look at. Try '|ylook bo|n',
    '|yl co|n' or '|yl sm-1|n'!

    ... Oh, and if you see database-ids like (#1245) by the name of objects,
    it's because you are playing with Builder-privileges or higher. Regular
    players will not see the numbers.

    Next try |ylook door|n.

"""

_DOOR_DESC_OUT = """
This is a solid wooden door leading to the outside of the cabin. Some
text is written on it:

    This is an |wexit|n. An exit is often named by its compass-direction like
    |weast|n, |wwest|n, |wnorthwest|n and so on, but it could be named
    anything, like this door. To use the exit, you just write its name. So by
    writing |ydoor|n you will leave the cabin.

"""

_DOOR_DESC_IN = """
This is a solid wooden door leading to the inside of the cabin. On
are some carved text:

    This exit leads back into the cabin. An exit is just like any object,
    so while has a name, it can also have aliases. To get back inside
    you can both write |ydoor|n but also |yin|n.

"""

_MEADOW_DESC = """
This is a lush meadow, just outside a cozy cabin. It's surrounded
by trees and sunlight filters down from a clear blue sky.

There is a |wstone|n here. Try looking at it!

"""

_STONE_DESC = """
This is a fist-sized stone covered in runes:

    To pick me up, use

    |yget stone|n

    You can see what you carry with the |yinventory|n (|yi|n).

    To drop me again, just write

    |ydrop stone|n

    Use |ynext|n when you are done exploring and want to
    continue with the tutorial.

"""


def _maintain_demo_room(caller, delete=False):
    """
    Handle the creation/cleanup of demo assets. We store them
    on the character and clean them when leaving the menu later.
    """
    # this is a tuple (room, obj)
    roomdata = caller.db.tutorial_world_demo_room_data

    if delete:
        if roomdata:
            # we delete directly for simplicity. We need to delete
            # in specific order to avoid deleting rooms moves
            # its contents to their default home-location
            prev_loc, room1, sign, room2, stone, door_out, door_in = roomdata
            caller.location = prev_loc
            sign.delete()
            stone.delete()
            door_out.delete()
            door_in.delete()
            room1.delete()
            room2.delete()
            del caller.db.tutorial_world_demo_room_data
    elif not roomdata:
        # create and describe the cabin and box
        room1 = create_object("evennia.objects.objects.DefaultRoom", key="A small, cozy cabin")
        room1.db.desc = _ROOM_DESC.lstrip()
        sign = create_object(
            "evennia.objects.objects.DefaultObject", key="small wooden sign", location=room1
        )
        sign.db.desc = _SIGN_DESC.strip()
        sign.locks.add("get:false()")
        sign.db.get_err_msg = "The sign is nailed to the wall. It's not budging."

        # create and describe the meadow and stone
        room2 = create_object("evennia.objects.objects.DefaultRoom", key="A lush summer meadow")
        room2.db.desc = _MEADOW_DESC.lstrip()
        stone = create_object(
            "evennia.objects.objects.DefaultObject", key="carved stone", location=room2
        )
        stone.db.desc = _STONE_DESC.strip()

        # make the linking exits
        door_out = create_object(
            "evennia.objects.objects.DefaultExit",
            key="Door",
            location=room1,
            destination=room2,
            locks=["get:false()"],
        )
        door_out.db.desc = _DOOR_DESC_OUT.strip()
        door_in = create_object(
            "evennia.objects.objects.DefaultExit",
            key="entrance to the cabin",
            aliases=["door", "in", "entrance"],
            location=room2,
            destination=room1,
            locks=["get:false()"],
        )
        door_in.db.desc = _DOOR_DESC_IN.strip()

        # store references for easy removal later
        caller.db.tutorial_world_demo_room_data = (
            caller.location,
            room1,
            sign,
            room2,
            stone,
            door_out,
            door_in,
        )
        # move caller into room
        caller.location = room1


class DemoCommandSetRoom(CmdSet):
    """
    Demo some general in-game commands command.
    """

    key = "Room Demo Set"
    priority = 2
    no_exits = False
    no_objs = False

    def at_cmdset_creation(self):
        from evennia import default_cmds

        self.add(default_cmds.CmdHelp())
        self.add(default_cmds.CmdLook())
        self.add(default_cmds.CmdGet())
        self.add(default_cmds.CmdDrop())
        self.add(default_cmds.CmdInventory())
        self.add(default_cmds.CmdExamine())
        self.add(default_cmds.CmdPy())


def goto_command_demo_room(caller, raw_string, **kwargs):
    """
    Setup and go to the demo-room node. Generates a little 2-room environment
    for testing out some commands.
    """
    _maintain_demo_room(caller)
    caller.cmdset.remove(DemoCommandSetHelp)
    caller.cmdset.remove(DemoCommandSetComms)
    caller.cmdset.add(DemoCommandSetRoom)
    return "command_demo_room"


def goto_cleanup_cmdsets(caller, raw_strings, **kwargs):
    """
    Cleanup all cmdsets.
    """
    caller.cmdset.remove(DemoCommandSetHelp)
    caller.cmdset.remove(DemoCommandSetComms)
    caller.cmdset.remove(DemoCommandSetRoom)
    return kwargs.get("gotonode")


# register all callables that can be used in the menu template

GOTO_CALLABLES = {
    "send_testing_tagged": send_testing_tagged,
    "do_nothing": do_nothing,
    "goto_command_demo_help": goto_command_demo_help,
    "goto_command_demo_comms": goto_command_demo_comms,
    "goto_command_demo_room": goto_command_demo_room,
    "goto_cleanup_cmdsets": goto_cleanup_cmdsets,
}


# Main menu definition

MENU_TEMPLATE = """

## NODE start

|g** Evennia introduction wizard **|n

If you feel lost you can learn some of the basics of how to play a text-based
game here. You can also learn a little about the system and how to find more
help. You can exit this tutorial-wizard at any time by entering '|yq|n' or '|yquit|n'.

Press |y<return>|n or write |ynext|n to step forward. Or select a number to jump to.

## OPTIONS

    1 (next);1;next;n: What is a MUD/MU*? -> about_muds
    2: About Evennia -> about_evennia
    3: Using the webclient -> using webclient
    4: The help command -> goto_command_demo_help()
    5: Communicating with others -> goto_command_demo_help(gotonode='talk on channels')
    6: Using colors -> goto_command_demo_comms(gotonode='testing_colors')
    7: Moving and exploring -> goto_command_demo_room()
    8: Conclusions & next steps-> conclusions
    >: about_muds

# ---------------------------------------------------------------------------------

## NODE about_muds

|g** About MUDs **|n

The term '|wMUD|n' stands for Multi-user-Dungeon or -Dimension. A MUD is
primarily played by inserting text |wcommands|n and getting text back.

MUDS were the |wprecursors|n to graphical MMORPG-style games like World of
Warcraft. While not as mainstream as they once were, comparing a text-game to a
graphical game is like comparing a book to a movie - it's just a different
experience altogether.

MUDs are |wdifferent|n from Interactive Fiction (IF) in that they are multiplayer
and usually have a consistent game world with many stories and protagonists
acting at the same time.

Like there are many different styles of graphical MMOs, there are |wmany
variations|n of MUDs: They can be slow-paced or fast. They can cover fantasy,
sci-fi, horror or other genres. They can allow PvP or not and be casual or
hardcore, strategic, tactical, turn-based or play in real-time.

Whereas 'MUD' is arguably the most well-known term, there are other terms
centered around particular game engines - such as MUSH, MOO, MUX, MUCK, LPMuds,
ROMs, Diku and others. Many people that played MUDs in the past used one of
these existing families of text game-servers, whether they knew it or not.

|cEvennia|n is a newer text game engine designed to emulate almost any existing
gaming style you like and possibly any new ones you can come up with!

## OPTIONS

    next;n: About Evennia -> about_evennia
    back to start;back;start;t: start
    >: about_evennia

# ---------------------------------------------------------------------------------

## NODE about_evennia

|g** About Evennia **|n

|cEvennia|n is a Python game engine for creating multiplayer online text-games
(aka MUDs, MUSHes, MUX, MOOs...). It is open-source and |wfree to use|n, also for
commercial projects (BSD license).

Out of the box, Evennia provides a |wworking, if empty game|n. Whereas you can play
via traditional telnet MUD-clients, the server runs your game's website and
offers a |wHTML5 webclient|n so that people can play your game in their browser
without downloading anything extra.

Evennia deliberately |wdoes not|n hard-code any game-specific things like
combat-systems, races, skills, etc. They would not match what just you wanted
anyway! Whereas we do have optional contribs with many examples, most of our
users use them as inspiration to make their own thing.

Evennia is developed entirely in |wPython|n, using modern developer practices.
The advantage of text is that even a solo developer or small team can
realistically make a competitive multiplayer game (as compared to a graphical
MMORPG which is one of the most expensive game types in existence to develop).
Many also use Evennia as a |wfun way to learn Python|n!

## OPTIONS

    next;n: Using the webclient -> using webclient
    back;b: About MUDs -> about_muds
    >: using webclient

# ---------------------------------------------------------------------------------

## NODE using webclient

|g** Using the Webclient **|n

|RNote: This is only relevant if you use Evennia's HTML5 web client. If you use a
third-party (telnet) mud-client, you can skip this section.|n

Evennia's web client is (for a local install) found by pointing your browser to

    |yhttp://localhost:4001/webclient|n

For a live example, the public Evennia demo can be found at

    |yhttps://demo.evennia.com/webclient|n

The web client starts out having two panes - the input-pane for entering commands
and the main window.

- Use |y<Return>|n (or click the arrow on the right) to send your input.
- Use |yCtrl + <up/down-arrow>|n to step back and forth in your command-history.
- Use |yCtrl + <Return>|n to add a new line to your input without sending.
(Cmd instead of Ctrl-key on Macs)

There is also some |wextra|n info to learn about customizing the webclient.

## OPTIONS

    extra: Customizing the webclient -> customizing the webclient
    next;n: Playing the game -> goto_command_demo_help()
    back;b: About Evennia -> about_evennia
    back to start;start: start
    >: goto_command_demo_help()

# ---------------------------------------------------------------------------------

# this is a dead-end 'leaf' of the menu

## NODE customizing the webclient

|g** Extra hints on customizing the Webclient **|n

|y1)|n The panes of the webclient can be resized and you can create additional panes.

- Press the little  plus (|w+|n) sign in the top left and a new tab will appear.
- Click and drag the tab and pull it far to the right and release when it creates two
  panes next to each other.

|y2)|n You can have certain server output only appear in certain panes.

- In your new rightmost pane, click the diamond (⯁) symbol at the top.
- Unselect everything and make sure to select "testing".
- Click the diamond again so the menu closes.
- Next, write "|ytest Hello world!|n". A test-text should appear in your rightmost pane!

|y3)|n You can customize general webclient settings by pressing the cogwheel in the upper
left corner. It allows to change things like font and if the client should play sound.

The "message routing" allows for rerouting text matching a certain regular expression (regex)
to a web client pane with a specific tag that you set yourself.

|y4)|n Close the right-hand pane with the |wX|n in the rop right corner.

## OPTIONS

    back;b: using webclient
    > test *: send tagged message to new pane -> send_testing_tagged()
    >: using webclient

# ---------------------------------------------------------------------------------

# we get here via goto_command_demo_help()

## NODE command_demo_help

|g** Playing the game **|n

Evennia has about |w90 default commands|n. They include useful administration/building
commands and a few limited "in-game" commands to serve as examples. They are intended
to be changed, extended and modified as you please.

First to try is |yhelp|n. This lists all commands |wcurrently|n available to you.

Use |yhelp <topic>|n to get specific help. Try |yhelp help|n to get help on using
the help command. For your game you could add help about your game, lore, rules etc
as well.

At the moment you probably only have |whelp|n and a |wchannel|n command
(the '<menu commands>' is just a placeholder to indicate you are using this menu).

We'll add more commands as we get to them in this tutorial - but we'll only
cover a small handful. Once you exit you'll find a lot more! Now let's try
those channels ...

## OPTIONS

    next;n: Talk on Channels -> talk on channels
    back;b: Using the webclient -> goto_cleanup_cmdsets(gotonode='using webclient')
    back to start;start: start
    >: talk on channels

# ---------------------------------------------------------------------------------

## NODE talk on channels

|g** Talk on Channels **|n

|wChannels|n are like in-game chatrooms. The |wChannel Names|n help-category
holds the names of the channels available to you right now. One such channel is
|wpublic|n. Use |yhelp public|n to see how to use it. Try it:

    |ypublic Hello World!|n

This will send a message to the |wpublic|n channel where everyone on that
channel can see it. If someone else is on your server, you may get a reply!

Evennia can link its in-game channels to external chat networks. This allows
you to talk with people not actually logged into the game.

## OPTIONS

    next;n: Talk to people in-game -> goto_command_demo_comms()
    back;b: Finding help -> goto_command_demo_help()
    back to start;start: start
    >: goto_command_demo_comms()

# ---------------------------------------------------------------------------------

# we get here via goto_command_demo_comms()

## NODE comms_demo_start

|g** Talk to people in-game **|n

You can also chat with people inside the game. If you try |yhelp|n now you'll
find you have a few more commands available for trying this out.

    |ysay Hello there!|n
    |y'Hello there!|n

|wsay|n is used to talk to people in the same location you are. Everyone in the
room will see what you have to say. A single quote |y'|n is a  convenient shortcut.

    |ypose smiles|n
    |y:smiles|n

|wpose|n (or |wemote|n) describes what you do to those nearby. This is a very simple
command by default, but it can be extended to much more complex parsing in order to
include other people/objects in the emote, reference things by a short-description etc.

## OPTIONS

    next;n: Paging people -> paging_people
    back;b: Talk on Channels -> goto_command_demo_help(gotonode='talk on channels')
    back to start;start: start
    >: paging_people

# ---------------------------------------------------------------------------------

## NODE paging_people

|g** Paging people **|n

Halfway between talking on a |wChannel|n and chatting in your current location
with |wsay|n and |wpose|n, you can also send private messages with |wpage|n:

    |ypage <name> Hello there!|n

Put your own name as |y<name>|n to page yourself as a test. Write just |ypage|n
to see your latest pages. This will also show you if anyone paged you while you
were offline.

## OPTIONS

    next;n: Using colors -> testing_colors
    back;b: Talk to people in-game -> comms_demo_start
    back to start;start: start
    >: testing_colors

# ---------------------------------------------------------------------------------

## NODE testing_colors

|g** U|rs|yi|gn|wg |c|yc|wo|rl|bo|gr|cs |g**|n

You can add color in your text by the help of tags. However, remember that not
everyone will see your colors - it depends on their client (and some use
screenreaders). Using color can also make text harder to read. So use it
sparingly.

To start coloring something |rred|n, add a ||r (red) marker and then
end with ||n (to go back to neutral/no-color):

    |ysay This is a ||rred||n text!
    say This is a ||Rdark red||n text!|n

You can also change the background:

    |ysay This is a ||[x||bblue text on a light-grey background!|n

There are 16 base colors and as many background colors (called ANSI colors). Some
clients also supports so-called Xterm256 which gives a total of 256 colors. These are
given as |w||rgb|n, where r, g, b are the components of red, green and blue from 0-5:

    |ysay This is ||050solid green!|n
    |ysay This is ||520an orange color!|n
    |ysay This is ||[005||555white on bright blue background!|n

If you don't see the expected colors from the above examples, it's because your
client does not support it - try out the Evennia webclient instead. To see all
color codes printed, try

    |ycolor ansi
    |ycolor xterm

## OPTIONS

    next;n: Moving and Exploring -> goto_command_demo_room()
    back;b: Paging people -> goto_command_demo_comms(gotonode='paging_people')
    back to start;start: start
    >: goto_command_demo_room()

# ---------------------------------------------------------------------------------

# we get here via goto_command_demo_room()

## NODE command_demo_room

|gMoving and Exploring|n

For exploring the game, a very important command is '|ylook|n'. It's also
abbreviated '|yl|n' since it's used so much. Looking displays/redisplays your
current location. You can also use it to look closer at items in the world. So
far in this tutorial, using 'look' would just redisplay the menu.

Try |ylook|n now. You have been quietly transported to a sunny cabin to look
around in. Explore a little and use |ynext|n when you are done.

## OPTIONS

    next;n: Conclusions -> conclusions
    back;b: Channel commands -> goto_command_demo_comms(gotonode='testing_colors')
    back to start;start: start
    >: conclusions

# ---------------------------------------------------------------------------------

## NODE conclusions

|gConclusions|n

That concludes this little quick-intro to using the base game commands of
Evennia. With this you should be able to continue exploring and also find help
if you get stuck!

Write |ynext|n to end this wizard. If you want there is also some |wextra|n info
for where to go beyond that.

## OPTIONS

    extra: Where to go next -> post scriptum
    next;next;n: End -> end
    back;b: Moving and Exploring -> goto_command_demo_room()
    back to start;start: start
    >: end

# ---------------------------------------------------------------------------------

## NODE post scriptum

|gWhere to next?|n

After playing through the tutorial-world quest, if you aim to make a game with
Evennia you are wise to take a look at the |wEvennia documentation|n at

    |yhttps://www.evennia.com/docs/latest|n

- You can start by trying to build some stuff by following the |wBuilder quick-start|n:

    |yhttps://www.evennia.com/docs/latest/Howtos/Beginner-Tutorial/Part1/Beginner-Tutorial-Building-Quickstart|n

- The tutorial-world may or may not be your cup of tea, but it does show off
  several |wuseful tools|n of Evennia. You may want to check out how it works:

    |yhttps://www.evennia.com/docs/latest/Howtos/Beginner-Tutorial/Part1/Beginner-Tutorial-Tutorial-World|n

- You can then continue looking through the |wTutorials|n and pick one that
  fits your level of understanding.

    |yhttps://www.evennia.com/docs/latest/Howtos/Howtos-Overview|n

- Make sure to |wjoin our forum|n and connect to our |wsupport chat|n! The
  Evennia community is very active and friendly and no question is too simple.
  You will often quickly get help. You can everything you need linked from

    |yhttps://www.evennia.com|n

# ---------------------------------------------------------------------------------

## OPTIONS

back: conclusions
>: conclusions


## NODE end

|gGood luck!|n

"""


# -------------------------------------------------------------------------------------------
#
# EvMenu implementation and access function
#
# -------------------------------------------------------------------------------------------


class TutorialEvMenu(EvMenu):
    """
    Custom EvMenu for displaying the intro-menu
    """

    def close_menu(self):
        """Custom cleanup actions when closing menu"""
        self.caller.cmdset.remove(DemoCommandSetHelp)
        self.caller.cmdset.remove(DemoCommandSetRoom)
        self.caller.cmdset.remove(DemoCommandSetComms)
        _maintain_demo_room(self.caller, delete=True)
        super().close_menu()
        if self.caller.account:
            self.caller.msg("Restoring permissions ...")
            self.caller.account.execute_cmd("unquell")

    def options_formatter(self, optionslist):
        navigation_keys = ("next", "back", "back to start")

        other = []
        navigation = []
        for key, desc in optionslist:
            if key in navigation_keys:
                desc = f" ({desc})" if desc else ""
                navigation.append(f"|lc{key}|lt|w{key}|n|le{desc}")
            else:
                other.append((key, desc))
        navigation = (
            (" " + " |W|||n ".join(navigation) + " |W|||n " + "|wQ|Wuit|n") if navigation else ""
        )
        other = super().options_formatter(other)
        sep = "\n\n" if navigation and other else ""

        return f"{navigation}{sep}{other}"


def init_menu(caller):
    """
    Call to initialize the menu.

    """
    menutree = parse_menu_template(caller, MENU_TEMPLATE, GOTO_CALLABLES)
    TutorialEvMenu(caller, menutree)



"""
Views are the functions that are called by different url endpoints. The Django
Rest Framework provides collections called 'ViewSets', which can generate a
number of views for the common CRUD operations.

"""
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.viewsets import ModelViewSet

from evennia.accounts.models import AccountDB
from evennia.help.models import HelpEntry
from evennia.objects.models import ObjectDB
from evennia.objects.objects import DefaultCharacter, DefaultExit, DefaultRoom
from evennia.scripts.models import ScriptDB
from evennia.web.api import filters, serializers
from evennia.web.api.permissions import EvenniaPermission


class GeneralViewSetMixin:
    """
    Mixin for both typeclass- and non-typeclass entities.

    """

    def get_serializer_class(self):
        """
        Allow different serializers for certain actions.

        """
        if self.action == "list":
            if hasattr(self, "list_serializer_class"):
                return self.list_serializer_class
        return self.serializer_class


class TypeclassViewSetMixin(GeneralViewSetMixin):
    """
    This mixin adds some shared functionality to each viewset of a typeclass. They all use the same
    permission classes and filter backend. You can override any of these in your own viewsets.

    The `set_atribute` action is  an example of a custom action added to a
    viewset. Based on the name of the method, it will create a default url_name
    (used for reversing) and url_path.  The 'pk' argument is automatically
    passed to this action because it has a url path of the format <object
    type>/:pk/set-attribute. The get_object method is automatically set in the
    expected viewset classes that will inherit this, using the pk that's passed
    along to retrieve the object.

    """

    # permission classes determine who is authorized to call the view
    permission_classes = [EvenniaPermission]
    # the filter backend allows for retrieval views to have filter arguments passed to it,
    # for example: mygame.com/api/objects?db_key=bob to find matches based on objects having a db_key of bob
    filter_backends = [DjangoFilterBackend]

    @action(detail=True, methods=["put", "post"])
    def set_attribute(self, request, pk=None):
        """
        This action will set an attribute if the db_value is defined, or remove
        it if no db_value is provided.

        """
        attr = serializers.AttributeSerializer(data=request.data)
        obj = self.get_object()
        if attr.is_valid(raise_exception=True):
            key = attr.validated_data["db_key"]
            value = attr.validated_data.get("db_value")
            category = attr.validated_data.get("db_category")
            attr_type = attr.validated_data.get("db_attrtype")
            if attr_type == "nick":
                handler = obj.nicks
            else:
                handler = obj.attributes
            if value:
                handler.add(key=key, value=value, category=category)
            else:
                handler.remove(key=key, category=category)
            return Response(
                serializers.AttributeSerializer(obj.db_attributes.all(), many=True).data,
                status=status.HTTP_200_OK,
            )
        return Response(attr.errors, status=status.HTTP_400_BAD_REQUEST)


class ObjectDBViewSet(TypeclassViewSetMixin, ModelViewSet):
    """
    The Object is the parent for all in-game entities that have a location
    (rooms, exits, characters etc).

    """

    # An example of a basic viewset for all ObjectDB instances. It declares the
    # serializer to use for both retrieving and changing/creating/deleting
    # instances. Serializers are similar to django forms, used for the
    # transmitting of data (typically json).

    serializer_class = serializers.ObjectDBSerializer
    queryset = ObjectDB.objects.all()
    filterset_class = filters.ObjectDBFilterSet
    list_serializer_class = serializers.ObjectListSerializer


class CharacterViewSet(ObjectDBViewSet):
    """
    Characters are a type of Object commonly used as player avatars in-game.

    """

    queryset = DefaultCharacter.objects.all_family()


class RoomViewSet(ObjectDBViewSet):
    """
    Rooms indicate discrete locations in-game.

    """

    queryset = DefaultRoom.objects.all_family()


class ExitViewSet(ObjectDBViewSet):
    """
    Exits are objects with a destination and allows for traversing from one
    location to another.

    """

    queryset = DefaultExit.objects.all_family()


class AccountDBViewSet(TypeclassViewSetMixin, ModelViewSet):
    """
    Accounts represent the players connected to the game

    """

    serializer_class = serializers.AccountSerializer
    queryset = AccountDB.objects.all()
    filterset_class = filters.AccountDBFilterSet
    list_serializer_class = serializers.AccountListSerializer


class ScriptDBViewSet(TypeclassViewSetMixin, ModelViewSet):
    """
    Scripts are meta-objects for storing system data, running timers etc. They
    have no in-game existence.

    """

    serializer_class = serializers.ScriptDBSerializer
    queryset = ScriptDB.objects.all()
    filterset_class = filters.ScriptDBFilterSet
    list_serializer_class = serializers.ScriptListSerializer


class HelpViewSet(GeneralViewSetMixin, ModelViewSet):
    """
    Database-stored help entries.
    Note that command auto-help and file-based help entries are not accessible this way.

    """

    serializer_class = serializers.HelpSerializer
    queryset = HelpEntry.objects.all()
    filterset_class = filters.HelpFilterSet
    list_serializer_class = serializers.HelpListSerializer



"""
This module handles initial database propagation, which is only run the first time the game starts.
It will create some default objects (notably give #1 its evennia-specific properties, and create the
Limbo room). It will also hooks, and then perform an initial restart.

Everything starts at handle_setup()
"""


import time

from django.conf import settings
from django.utils.translation import gettext as _

import evennia
from evennia.accounts.models import AccountDB
from evennia.server.models import ServerConfig
from evennia.utils import create, logger

ERROR_NO_SUPERUSER = """
    No superuser exists yet. The superuser is the 'owner' account on
    the Evennia server. Create a new superuser using the command

       evennia createsuperuser

    Follow the prompts, then restart the server.
    """


LIMBO_DESC = _(
    """
Welcome to your new |wEvennia|n-based game! Visit https://www.evennia.com if you need
help, want to contribute, report issues or just join the community.

As a privileged user, write |wbatchcommand tutorial_world.build|n to build
tutorial content. Once built, try |wintro|n for starting help and |wtutorial|n to
play the demo game.
"""
)


WARNING_POSTGRESQL_FIX = """
    PostgreSQL-psycopg2 compatibility fix:
    The in-game channels {chan1}, {chan2} and {chan3} were created,
    but the superuser was not yet connected to them. Please use in
    game commands to connect Account #1 to those channels when first
    logging in.
"""


def _get_superuser_account():
    """
    Get the superuser (created at the command line) and don't take no for an answer.

    Returns:
        Account: The first superuser (User #1).

    Raises:
        AccountDB.DoesNotExist: If the superuser couldn't be found.

    """
    try:
        superuser = AccountDB.objects.get(id=1)
    except AccountDB.DoesNotExist:
        raise AccountDB.DoesNotExist(ERROR_NO_SUPERUSER)
    return superuser


def create_objects():
    """
    Creates the #1 account and Limbo room.

    """

    logger.log_info("Initial setup: Creating objects (Account #1 and Limbo room) ...")

    # Set the initial User's account object's username on the #1 object.
    # This object is pure django and only holds name, email and password.
    superuser = _get_superuser_account()
    from evennia.objects.models import ObjectDB

    # Create an Account 'user profile' object to hold eventual
    # mud-specific settings for the AccountDB object.
    account_typeclass = settings.BASE_ACCOUNT_TYPECLASS

    # run all creation hooks on superuser (we must do so manually
    # since the manage.py command does not)
    superuser.swap_typeclass(account_typeclass, clean_attributes=True)
    superuser.basetype_setup()
    superuser.at_account_creation()
    superuser.locks.add(
        "examine:perm(Developer);edit:false();delete:false();boot:false();msg:all()"
    )
    # this is necessary for quelling to work correctly.
    superuser.permissions.add("Developer")

    # Limbo is the default "nowhere" starting room

    # Create the in-game god-character for account #1 and set
    # it to exist in Limbo.
    try:
        superuser_character = ObjectDB.objects.get(id=1)
    except ObjectDB.DoesNotExist:
        superuser_character, errors = superuser.create_character(
            key=superuser.username, nohome=True, description=_("This is User #1.")
        )
        if errors:
            raise Exception(str(errors))

    superuser_character.locks.add(
        "examine:perm(Developer);edit:false();delete:false();boot:false();msg:all();puppet:false()"
    )
    # we set this low so that quelling is more useful
    superuser_character.permissions.add("Developer")
    superuser_character.save()

    superuser.attributes.add("_first_login", True)
    superuser.attributes.add("_last_puppet", superuser_character)

    room_typeclass = settings.BASE_ROOM_TYPECLASS
    try:
        limbo_obj = ObjectDB.objects.get(id=2)
    except ObjectDB.DoesNotExist:
        limbo_obj = create.create_object(room_typeclass, _("Limbo"), nohome=True)

    limbo_obj.db_typeclass_path = room_typeclass
    limbo_obj.db.desc = LIMBO_DESC.strip()
    limbo_obj.save()

    # Now that Limbo exists, try to set the user up in Limbo (unless
    # the creation hooks already fixed this).
    if not superuser_character.location:
        superuser_character.location = limbo_obj
    if not superuser_character.home:
        superuser_character.home = limbo_obj


def at_initial_setup():
    """
    Custom hook for users to overload some or all parts of the initial
    setup. Called very last in the sequence. It tries to import and
    srun a module settings.AT_INITIAL_SETUP_HOOK_MODULE and will fail
    silently if this does not exist or fails to load.

    """
    modname = settings.AT_INITIAL_SETUP_HOOK_MODULE
    if not modname:
        return
    try:
        mod = __import__(modname, fromlist=[None])
    except (ImportError, ValueError):
        return
    logger.log_info("Initial setup: Running at_initial_setup() hook.")
    if mod.__dict__.get("at_initial_setup", None):
        mod.at_initial_setup()


def collectstatic():
    """
    Run collectstatic to make sure all web assets are loaded.

    """
    from django.core.management import call_command

    logger.log_info("Initial setup: Gathering static resources using 'collectstatic'")
    call_command("collectstatic", "--noinput")


def reset_server():
    """
    We end the initialization by resetting the server. This makes sure
    the first login is the same as all the following ones,
    particularly it cleans all caches for the special objects.  It
    also checks so the warm-reset mechanism works as it should.

    """
    if settings.TEST_ENVIRONMENT:
        return
    ServerConfig.objects.conf("server_epoch", time.time())

    logger.log_info("Initial setup complete. Restarting Server once.")
    evennia.SESSION_HANDLER.portal_reset_server()


def handle_setup(last_step=None):
    """
    Main logic for the module. It allows for restarting the
    initialization at any point if one of the modules should crash.

    Args:
        last_step (str, None): The last stored successful step, for starting
            over on errors. None if starting from scratch. If this is 'done',
            the function will exit immediately.

    """
    # setup sequence
    setup_sequence = {
        "create_objects": create_objects,
        "at_initial_setup": at_initial_setup,
        "collectstatic": collectstatic,
        "done": reset_server,
    }

    if last_step in ("done", -1):
        # this means we don't need to handle setup since
        # it already ran sucessfully once. -1 is the legacy
        # value for existing databases.
        return

    # determine the sequence so we can skip ahead
    steps = list(setup_sequence)
    steps = steps[steps.index(last_step) + 1 if last_step is not None else 0 :]

    # step through queue from last completed function. Once completed,
    # the 'done' key should be set.
    for stepname in steps:
        try:
            setup_sequence[stepname]()
        except Exception:
            # we re-raise to make sure to stop startup
            raise
        else:
            # save the step
            ServerConfig.objects.conf("last_initial_setup_step", stepname)
            if stepname == "done":
                # always exit on 'done'
                break



# Generated by Django 3.2.3 on 2021-05-30 18:18

from django.db import migrations


def update_help_entries(apps, schema_editor):
    """
    Change all help-entry files that use view: locks to read: locks
    (read: was not used before and view: was previously what read is now).

    """
    HelpEntry = apps.get_model("help", "HelpEntry")
    for help_entry in HelpEntry.objects.all():
        lock_storage = help_entry.db_lock_storage
        lock_storage = dict(
            lstring.split(":", 1) if ":" in lstring else (lstring, "")
            for lstring in str(lock_storage).split(";")
        )
        if "read" in lock_storage:
            # already in place - skip
            continue
        if "view" in lock_storage:
            lock_storage["read"] = lock_storage.pop("view")
        lock_storage = ";".join(f"{typ}:{lock}" for typ, lock in lock_storage.items())
        help_entry.db_lock_storage = lock_storage
        help_entry.save(update_fields=["db_lock_storage"])


class Migration(migrations.Migration):

    dependencies = [
        ("help", "0004_auto_20210520_2137"),
    ]

    operations = [migrations.RunPython(update_help_entries)]



"""
Evennia Multidescer

Contrib - Griatch 2016

A "multidescer" is a concept from the MUSH world. It allows for
creating, managing and switching between multiple character
descriptions. This multidescer will not require any changes to the
Character class, rather it will use the `multidescs` Attribute (a
list) and create it if it does not exist.

This contrib also works well together with the rpsystem contrib (which
also adds the short descriptions and the `sdesc` command).

Installation:

Edit `mygame/commands/default_cmdsets.py` and add
`from evennia.contrib.game_systems.multidescer import CmdMultiDesc` to the top.

Next, look up the `at_cmdset_create` method of the `CharacterCmdSet`
class and add a line `self.add(CmdMultiDesc())` to the end
of it.

Reload the server and you should have the +desc command available (it
will replace the default `desc` command).

"""
import re

from evennia import default_cmds
from evennia.utils.eveditor import EvEditor
from evennia.utils.utils import crop

# regex for the set functionality
_RE_KEYS = re.compile(r"([\w\s]+)(?:\+*?)", re.U + re.I)


# Helper functions for the Command


class DescValidateError(ValueError):
    "Used for tracebacks from desc systems"
    pass


def _update_store(caller, key=None, desc=None, delete=False, swapkey=None):
    """
    Helper function for updating the database store.

    Args:
        caller (Object): The caller of the command.
        key (str): Description identifier
        desc (str): Description text.
        delete (bool): Delete given key.
        swapkey (str): Swap list positions of `key` and this key.

    """
    if not caller.db.multidesc:
        # initialize the multidesc attribute
        caller.db.multidesc = [("caller", caller.db.desc or "")]
    if not key:
        return
    lokey = key.lower()
    match = [ind for ind, tup in enumerate(caller.db.multidesc) if tup[0] == lokey]
    if match:
        idesc = match[0]
        if delete:
            # delete entry
            del caller.db.multidesc[idesc]
        elif swapkey:
            # swap positions
            loswapkey = swapkey.lower()
            swapmatch = [ind for ind, tup in enumerate(caller.db.multidesc) if tup[0] == loswapkey]
            if swapmatch:
                iswap = swapmatch[0]
                if idesc == iswap:
                    raise DescValidateError("Swapping a key with itself does nothing.")
                temp = caller.db.multidesc[idesc]
                caller.db.multidesc[idesc] = caller.db.multidesc[iswap]
                caller.db.multidesc[iswap] = temp
            else:
                raise DescValidateError("Description key '|w%s|n' not found." % swapkey)
        elif desc:
            # update in-place
            caller.db.multidesc[idesc] = (lokey, desc)
        else:
            raise DescValidateError("No description was set.")
    else:
        # no matching key
        if delete or swapkey:
            raise DescValidateError("Description key '|w%s|n' not found." % key)
        elif desc:
            # insert new at the top of the stack
            caller.db.multidesc.insert(0, (lokey, desc))
        else:
            raise DescValidateError("No description was set.")


# eveditor save/load/quit functions


def _save_editor(caller, buffer):
    "Called when the editor saves its contents"
    key = caller.db._multidesc_editkey
    _update_store(caller, key, buffer)
    caller.msg("Saved description to key '%s'." % key)
    return True


def _load_editor(caller):
    "Called when the editor loads contents"
    key = caller.db._multidesc_editkey
    match = [ind for ind, tup in enumerate(caller.db.multidesc) if tup[0] == key]
    if match:
        return caller.db.multidesc[match[0]][1]
    return ""


def _quit_editor(caller):
    "Called when the editor quits"
    del caller.db._multidesc_editkey
    caller.msg("Exited editor.")


# The actual command class


class CmdMultiDesc(default_cmds.MuxCommand):
    """
    Manage multiple descriptions

    Usage:
        +desc [key]                - show current desc desc with <key>
        +desc <key> = <text>       - add/replace desc with <key>
        +desc/list                 - list descriptions (abbreviated)
        +desc/list/full            - list descriptions (full texts)
        +desc/edit <key>           - add/edit desc <key> in line editor
        +desc/del <key>            - delete desc <key>
        +desc/swap <key1>-<key2>   - swap positions of <key1> and <key2> in list
        +desc/set <key> [+key+...] - set desc as default or combine multiple descs

    Notes:
        When combining multiple descs with +desc/set <key> + <key2> + ...,
        any keys not matching an actual description will be inserted
        as plain text. Use e.g. ansi line break ||/ to add a new
        paragraph and + + or ansi space ||_ to add extra whitespace.

    """

    key = "+desc"
    aliases = ["desc"]
    locks = "cmd:all()"
    help_category = "General"

    def func(self):
        """
        Implements the multidescer.  We will use `db.desc` for the
        description in use and `db.multidesc` to store all descriptions.
        """

        caller = self.caller
        args = self.args.strip()
        switches = self.switches

        try:
            if "list" in switches or "all" in switches:
                # list all stored descriptions, either in full or cropped.
                # Note that we list starting from 1, not from 0.
                _update_store(caller)
                do_crop = "full" not in switches
                if do_crop:
                    outtext = [
                        "|w%s:|n %s" % (key, crop(desc)) for key, desc in caller.db.multidesc
                    ]
                else:
                    outtext = [
                        "\n|w%s:|n|n\n%s\n%s" % (key, "-" * (len(key) + 1), desc)
                        for key, desc in caller.db.multidesc
                    ]

                caller.msg("|wStored descs:|n\n" + "\n".join(outtext))
                return

            elif "edit" in switches:
                # Use the eveditor to edit/create the named description
                if not args:
                    caller.msg("Usage: %s/edit key" % self.key)
                    return

                # this is used by the editor to know what to edit; it's deleted automatically
                caller.db._multidesc_editkey = args
                # start the editor
                EvEditor(
                    caller,
                    loadfunc=_load_editor,
                    savefunc=_save_editor,
                    quitfunc=_quit_editor,
                    key="multidesc editor",
                    persistent=True,
                )

            elif "delete" in switches or "del" in switches:
                # delete a multidesc entry.
                if not args:
                    caller.msg("Usage: %s/delete key" % self.key)
                    return
                _update_store(caller, args, delete=True)
                caller.msg("Deleted description with key '%s'." % args)

            elif "swap" in switches or "switch" in switches or "reorder" in switches:
                # Reorder list by swapping two entries. We expect numbers starting from 1
                keys = [arg for arg in args.split("-", 1)]
                if not len(keys) == 2:
                    caller.msg("Usage: %s/swap key1-key2" % self.key)
                    return
                key1, key2 = keys
                # perform the swap
                _update_store(caller, key1, swapkey=key2)
                caller.msg("Swapped descs '%s' and '%s'." % (key1, key2))

            elif "set" in switches:
                # switches one (or more) of the multidescs to be the "active" description
                _update_store(caller)
                if not args:
                    caller.msg("Usage: %s/set key [+ key2 + key3 + ...]" % self.key)
                    return
                new_desc = []
                multidesc = caller.db.multidesc
                for key in args.split("+"):
                    notfound = True
                    lokey = key.strip().lower()
                    for mkey, desc in multidesc:
                        if lokey == mkey:
                            new_desc.append(desc)
                            notfound = False
                            continue
                    if notfound:
                        # if we get here, there is no desc match, we add it as a normal string
                        new_desc.append(key)
                new_desc = "".join(new_desc)
                caller.db.desc = new_desc
                caller.msg("%s\n\n|wThe above was set as the current description.|n" % new_desc)

            elif self.rhs or "add" in switches:
                # add text directly to a new entry or an existing one.
                if not (self.lhs and self.rhs):
                    caller.msg("Usage: %s/add key = description" % self.key)
                    return
                key, desc = self.lhs, self.rhs
                _update_store(caller, key, desc)
                caller.msg("Stored description '%s': \"%s\"" % (key, crop(desc)))

            else:
                # display the current description or a numbered description
                _update_store(caller)
                if args:
                    key = args.lower()
                    multidesc = caller.db.multidesc
                    for mkey, desc in multidesc:
                        if key == mkey:
                            caller.msg("|wDecsription %s:|n\n%s" % (key, desc))
                            return
                    caller.msg("Description key '%s' not found." % key)
                else:
                    caller.msg("|wCurrent desc:|n\n%s" % caller.db.desc)

        except DescValidateError as err:
            # This is triggered by _key_to_index
            caller.msg(err)



"""
Test EvAdventure combat.

"""

from unittest.mock import Mock, call, patch

from evennia.utils import create
from evennia.utils.ansi import strip_ansi
from evennia.utils.test_resources import (
    BaseEvenniaTest,
    EvenniaCommandTestMixin,
    EvenniaTestCase,
)

from .. import combat_base, combat_turnbased, combat_twitch
from ..characters import EvAdventureCharacter
from ..enums import Ability, WieldLocation
from ..npcs import EvAdventureMob
from ..objects import EvAdventureConsumable, EvAdventureRunestone, EvAdventureWeapon
from ..rooms import EvAdventureRoom


class _CombatTestBase(EvenniaTestCase):
    """
    Set up common entities for testing combat:

    - `location` (key=testroom)
    - `combatant` (key=testchar)
    - `target` (key=testmonster)`

    We also mock the `.msg` method of both `combatant` and `target` so we can
    see what was sent.

    """

    def setUp(self):
        self.location = create.create_object(EvAdventureRoom, key="testroom")
        self.combatant = create.create_object(
            EvAdventureCharacter, key="testchar", location=self.location
        )

        self.location.allow_combat = True
        self.location.allow_death = True

        self.target = create.create_object(
            EvAdventureMob,
            key="testmonster",
            location=self.location,
            attributes=(("is_idle", True),),
        )

        # mock the msg so we can check what they were sent later
        self.combatant.msg = Mock()
        self.target.msg = Mock()


class TestEvAdventureCombatBaseHandler(_CombatTestBase):
    """
    Test the base functionality of the base combat handler.

    """

    def setUp(self):
        """This also tests the `get_or_create_combathandler` classfunc"""
        super().setUp()
        self.combathandler = combat_base.EvAdventureCombatBaseHandler.get_or_create_combathandler(
            self.location, key="combathandler"
        )

    def test_combathandler_msg(self):
        """Test sending messages to all in handler"""

        self.location.msg_contents = Mock()

        self.combathandler.msg("test_message")

        self.location.msg_contents.assert_called_with(
            "test_message",
            exclude=[],
            from_obj=None,
            mapping={"testchar": self.combatant, "testmonster": self.target},
        )

    def test_get_combat_summary(self):
        """Test combat summary"""

        self.combathandler.get_sides = Mock(return_value=([self.combatant], [self.target]))

        # as seen from one side
        result = str(self.combathandler.get_combat_summary(self.combatant))

        self.assertEqual(
            strip_ansi(result),
            " testchar (Perfect)  vs  testmonster (Perfect) ",
        )

        # as seen from other side
        self.combathandler.get_sides = Mock(return_value=([self.target], [self.combatant]))
        result = str(self.combathandler.get_combat_summary(self.target))

        self.assertEqual(
            strip_ansi(result),
            " testmonster (Perfect)  vs  testchar (Perfect) ",
        )


class TestCombatActionsBase(_CombatTestBase):
    """
    A class for testing all subclasses of CombatAction in combat_base.py

    """

    def setUp(self):
        super().setUp()
        self.combathandler = combat_base.EvAdventureCombatBaseHandler.get_or_create_combathandler(
            self.location, key="combathandler"
        )
        # we need to mock all NotImplemented methods
        self.combathandler.get_sides = Mock(return_value=([], [self.target]))
        self.combathandler.give_advantage = Mock()
        self.combathandler.give_disadvantage = Mock()
        self.combathandler.remove_advantage = Mock()
        self.combathandler.remove_disadvantage = Mock()
        self.combathandler.get_advantage = Mock()
        self.combathandler.get_disadvantage = Mock()
        self.combathandler.has_advantage = Mock()
        self.combathandler.has_disadvantage = Mock()
        self.combathandler.queue_action = Mock()

    def test_base_action(self):
        action = combat_base.CombatAction(
            self.combathandler, self.combatant, {"key": "hold", "foo": "bar"}
        )
        self.assertEqual(action.key, "hold")
        self.assertEqual(action.foo, "bar")
        self.assertEqual(action.combathandler, self.combathandler)
        self.assertEqual(action.combatant, self.combatant)

    @patch("evennia.contrib.tutorials.evadventure.combat_base.rules.randint")
    def test_attack__miss(self, mock_randint):
        actiondict = {"key": "attack", "target": self.target}

        mock_randint.return_value = 8  # target has default armor 11, so 8+1 str will miss
        action = combat_base.CombatActionAttack(self.combathandler, self.combatant, actiondict)
        action.execute()
        self.assertEqual(self.target.hp, 4)

    @patch("evennia.contrib.tutorials.evadventure.combat_base.rules.randint")
    def test_attack__success(self, mock_randint):
        actiondict = {"key": "attack", "target": self.target}

        mock_randint.return_value = 11  # 11 + 1 str will hit beat armor 11
        self.target.hp = 20
        action = combat_base.CombatActionAttack(self.combathandler, self.combatant, actiondict)
        action.execute()
        self.assertEqual(self.target.hp, 9)

    @patch("evennia.contrib.tutorials.evadventure.combat_base.rules.randint")
    def test_stunt_fail(self, mock_randint):
        action_dict = {
            "key": "stunt",
            "recipient": self.combatant,
            "target": self.target,
            "advantage": True,
            "stunt_type": Ability.STR,
            "defense_type": Ability.DEX,
        }
        mock_randint.return_value = 8  # fails 8+1 dex vs DEX 11 defence
        action = combat_base.CombatActionStunt(self.combathandler, self.combatant, action_dict)
        action.execute()
        self.combathandler.give_advantage.assert_not_called()

    @patch("evennia.contrib.tutorials.evadventure.combat_base.rules.randint")
    def test_stunt_advantage__success(self, mock_randint):
        action_dict = {
            "key": "stunt",
            "recipient": self.combatant,
            "target": self.target,
            "advantage": True,
            "stunt_type": Ability.STR,
            "defense_type": Ability.DEX,
        }
        mock_randint.return_value = 11  # 11+1 dex vs DEX 11 defence is success
        action = combat_base.CombatActionStunt(self.combathandler, self.combatant, action_dict)
        action.execute()
        self.combathandler.give_advantage.assert_called_with(self.combatant, self.target)

    @patch("evennia.contrib.tutorials.evadventure.combat_base.rules.randint")
    def test_stunt_disadvantage__success(self, mock_randint):
        action_dict = {
            "key": "stunt",
            "recipient": self.target,
            "target": self.combatant,
            "advantage": False,
            "stunt_type": Ability.STR,
            "defense_type": Ability.DEX,
        }
        mock_randint.return_value = 11  # 11+1 dex vs DEX 11 defence is success
        action = combat_base.CombatActionStunt(self.combathandler, self.combatant, action_dict)
        action.execute()
        self.combathandler.give_disadvantage.assert_called_with(self.target, self.combatant)

    def test_use_item(self):
        """
        Use up a potion during combat.

        """
        item = create.create_object(
            EvAdventureConsumable, key="Healing potion", attributes=[("uses", 2)]
        )

        item.use = Mock()

        action_dict = {
            "key": "use",
            "item": item,
            "target": self.target,
        }

        self.assertEqual(item.uses, 2)
        action = combat_base.CombatActionUseItem(self.combathandler, self.combatant, action_dict)
        action.execute()
        self.assertEqual(item.uses, 1)
        action.execute()
        self.assertEqual(item.pk, None)  # deleted, it was used up

    def test_swap_wielded_weapon_or_spell(self):
        """
        First draw a weapon (from empty fists), then swap that out to another weapon, then
        swap to a spell rune.

        """
        sword = create.create_object(EvAdventureWeapon, key="sword")
        zweihander = create.create_object(
            EvAdventureWeapon,
            key="zweihander",
            attributes=(("inventory_use_slot", WieldLocation.TWO_HANDS),),
        )
        runestone = create.create_object(EvAdventureRunestone, key="ice rune")

        # check hands are empty
        self.assertEqual(self.combatant.weapon.key, "Bare hands")
        self.assertEqual(self.combatant.equipment.slots[WieldLocation.WEAPON_HAND], None)
        self.assertEqual(self.combatant.equipment.slots[WieldLocation.TWO_HANDS], None)

        # swap to sword

        actiondict = {"key": "wield", "item": sword}

        action = combat_base.CombatActionWield(self.combathandler, self.combatant, actiondict)
        action.execute()

        self.assertEqual(self.combatant.weapon, sword)
        self.assertEqual(self.combatant.equipment.slots[WieldLocation.WEAPON_HAND], sword)
        self.assertEqual(self.combatant.equipment.slots[WieldLocation.TWO_HANDS], None)

        # swap to zweihander (two-handed sword)
        actiondict["item"] = zweihander

        action = combat_base.CombatActionWield(self.combathandler, self.combatant, actiondict)
        action.execute()

        self.assertEqual(self.combatant.weapon, zweihander)
        self.assertEqual(self.combatant.equipment.slots[WieldLocation.WEAPON_HAND], None)
        self.assertEqual(self.combatant.equipment.slots[WieldLocation.TWO_HANDS], zweihander)

        # swap to runestone (also using two hands)
        actiondict["item"] = runestone

        action = combat_base.CombatActionWield(self.combathandler, self.combatant, actiondict)
        action.execute()

        self.assertEqual(self.combatant.weapon, runestone)
        self.assertEqual(self.combatant.equipment.slots[WieldLocation.WEAPON_HAND], None)
        self.assertEqual(self.combatant.equipment.slots[WieldLocation.TWO_HANDS], runestone)

        # swap back to normal one-handed sword
        actiondict["item"] = sword

        action = combat_base.CombatActionWield(self.combathandler, self.combatant, actiondict)
        action.execute()

        self.assertEqual(self.combatant.weapon, sword)
        self.assertEqual(self.combatant.equipment.slots[WieldLocation.WEAPON_HAND], sword)
        self.assertEqual(self.combatant.equipment.slots[WieldLocation.TWO_HANDS], None)


class EvAdventureTurnbasedCombatHandlerTest(_CombatTestBase):
    """
    Test methods on the turn-based combat handler and actions

    """

    maxDiff = None

    # make sure to mock away all time-keeping elements
    @patch(
        (
            "evennia.contrib.tutorials.evadventure."
            "combat_turnbased.EvAdventureTurnbasedCombatHandler.interval"
        ),
        new=-1,
    )
    def setUp(self):
        super().setUp()
        # add target to combat
        self.combathandler = (
            combat_turnbased.EvAdventureTurnbasedCombatHandler.get_or_create_combathandler(
                self.location, key="combathandler"
            )
        )
        self.combathandler.add_combatant(self.combatant)
        self.combathandler.add_combatant(self.target)

    def _get_action(self, action_dict={"key": "hold"}):
        action_class = self.combathandler.action_classes[action_dict["key"]]
        return action_class(self.combathandler, self.combatant, action_dict)

    def _run_actions(
        self, action_dict, action_dict2={"key": "hold"}, combatant_msg=None, target_msg=None
    ):
        """
        Helper method to run an action and check so combatant saw the expected message.
        """
        self.combathandler.queue_action(self.combatant, action_dict)
        self.combathandler.queue_action(self.target, action_dict2)
        self.combathandler.at_repeat()
        if combatant_msg is not None:
            # this works because we mock combatant.msg in SetUp
            self.combatant.msg.assert_called_with(combatant_msg)
        if target_msg is not None:
            # this works because we mock target.msg in SetUp
            self.combatant.msg.assert_called_with(target_msg)

    def test_combatanthandler_setup(self):
        """Testing all is set up correctly in the combathandler"""

        chandler = self.combathandler
        self.assertEqual(
            dict(chandler.combatants),
            {self.combatant: {"key": "hold"}, self.target: {"key": "hold"}},
        )
        self.assertEqual(
            dict(chandler.action_classes),
            {
                "hold": combat_turnbased.CombatActionHold,
                "attack": combat_turnbased.CombatActionAttack,
                "stunt": combat_turnbased.CombatActionStunt,
                "use": combat_turnbased.CombatActionUseItem,
                "wield": combat_turnbased.CombatActionWield,
                "flee": combat_turnbased.CombatActionFlee,
            },
        )
        self.assertEqual(chandler.flee_timeout, 1)
        self.assertEqual(dict(chandler.advantage_matrix), {})
        self.assertEqual(dict(chandler.disadvantage_matrix), {})
        self.assertEqual(dict(chandler.fleeing_combatants), {})
        self.assertEqual(dict(chandler.defeated_combatants), {})

    def test_remove_combatant(self):
        """Remove a combatant."""

        self.combathandler.remove_combatant(self.target)
        self.assertEqual(dict(self.combathandler.combatants), {self.combatant: {"key": "hold"}})

    def test_stop_combat(self):
        """Stopping combat, making sure combathandler is deleted."""

        self.combathandler.stop_combat()
        self.assertIsNone(self.combathandler.pk)

    def test_get_sides(self):
        """Getting the sides of combat"""

        combatant2 = create.create_object(
            EvAdventureCharacter, key="testchar2", location=self.location
        )
        target2 = create.create_object(
            EvAdventureMob,
            key="testmonster2",
            location=self.location,
            attributes=(("is_idle", True),),
        )
        self.combathandler.add_combatant(combatant2)
        self.combathandler.add_combatant(target2)

        # allies to combatant
        allies, enemies = self.combathandler.get_sides(self.combatant)
        self.assertEqual((allies, enemies), ([self.combatant, combatant2], [self.target, target2]))

        # allies to monster
        allies, enemies = self.combathandler.get_sides(self.target)
        self.assertEqual((allies, enemies), ([self.target, target2], [self.combatant, combatant2]))

    def test_queue_and_execute_action(self):
        """Queue actions and execute"""

        hold = {"key": "hold"}

        self.combathandler.queue_action(self.combatant, hold)
        self.assertEqual(
            dict(self.combathandler.combatants),
            {self.combatant: {"key": "hold"}, self.target: {"key": "hold"}},
        )

        mock_action = Mock()
        self.combathandler.action_classes["hold"] = Mock(return_value=mock_action)

        self.combathandler.execute_next_action(self.combatant)

        self.combathandler.action_classes["hold"].assert_called_with(
            self.combathandler, self.combatant, hold
        )
        mock_action.execute.assert_called_once()

    def test_execute_full_turn(self):
        """Run a full (passive) turn"""

        hold = {"key": "hold"}

        self.combathandler.queue_action(self.combatant, hold)
        self.combathandler.queue_action(self.target, hold)

        self.combathandler.execute_next_action = Mock()

        self.combathandler.at_repeat()

        self.combathandler.execute_next_action.assert_has_calls(
            [call(self.combatant), call(self.target)], any_order=True
        )

    def test_action__action_ticks_turn(self):
        """Test that action execution ticks turns"""

        actiondict = {"key": "hold"}
        self._run_actions(actiondict, actiondict)
        self.assertEqual(self.combathandler.turn, 1)

        self.combatant.msg.assert_not_called()

    @patch("evennia.contrib.tutorials.evadventure.combat_base.rules.randint")
    def test_attack__success__kill(self, mock_randint):
        """Test that the combathandler is deleted once there are no more enemies"""
        actiondict = {"key": "attack", "target": self.target}

        mock_randint.return_value = 11  # 11 + 1 str will hit beat armor 11
        self._run_actions(actiondict)
        self.assertEqual(self.target.hp, -7)
        # after this the combat is over
        self.assertIsNone(self.combathandler.pk)

    @patch("evennia.contrib.tutorials.evadventure.combat_base.rules.randint")
    def test_stunt_fail(self, mock_randint):
        action_dict = {
            "key": "stunt",
            "recipient": self.combatant,
            "target": self.target,
            "advantage": True,
            "stunt_type": Ability.STR,
            "defense_type": Ability.DEX,
        }
        mock_randint.return_value = 8  # fails 8+1 dex vs DEX 11 defence
        self._run_actions(action_dict)
        self.assertEqual(self.combathandler.advantage_matrix[self.combatant], {})
        self.assertEqual(self.combathandler.disadvantage_matrix[self.combatant], {})

    @patch("evennia.contrib.tutorials.evadventure.combat_base.rules.randint")
    def test_stunt_advantage__success(self, mock_randint):
        """Test so the advantage matrix is updated correctly"""
        action_dict = {
            "key": "stunt",
            "recipient": self.combatant,
            "target": self.target,
            "advantage": True,
            "stunt_type": Ability.STR,
            "defense_type": Ability.DEX,
        }
        mock_randint.return_value = 11  # 11+1 dex vs DEX 11 defence is success
        self._run_actions(action_dict)
        self.assertEqual(
            bool(self.combathandler.advantage_matrix[self.combatant][self.target]), True
        )

    @patch("evennia.contrib.tutorials.evadventure.combat_base.rules.randint")
    def test_stunt_disadvantage__success(self, mock_randint):
        """Test so the disadvantage matrix is updated correctly"""
        action_dict = {
            "key": "stunt",
            "recipient": self.target,
            "target": self.combatant,
            "advantage": False,
            "stunt_type": Ability.STR,
            "defense_type": Ability.DEX,
        }
        mock_randint.return_value = 11  # 11+1 dex vs DEX 11 defence is success
        self._run_actions(action_dict)
        self.assertEqual(
            bool(self.combathandler.disadvantage_matrix[self.target][self.combatant]), True
        )

    def test_flee__success(self):
        """
        Test fleeing twice, leading to leaving combat.

        """

        self.assertEqual(self.combathandler.turn, 0)
        action_dict = {"key": "flee", "repeat": True}

        # first flee records the fleeing state
        self.combathandler.flee_timeout = 2  # to make sure
        self._run_actions(action_dict)
        self.assertEqual(self.combathandler.turn, 1)
        self.assertEqual(self.combathandler.fleeing_combatants[self.combatant], 1)

        # action_dict should still be in place due to repeat
        self.assertEqual(self.combathandler.combatants[self.combatant], action_dict)

        self.combatant.msg.assert_called_with(
            text=(
                "You retreat, being exposed to attack while doing so (will escape in 1 turn).",
                {},
            ),
            from_obj=self.combatant,
        )
        # Check that enemies have advantage against you now
        action = combat_turnbased.CombatAction(self.combathandler, self.target, {"key": "hold"})
        self.assertTrue(action.combathandler.has_advantage(self.target, self.combatant))

        # second flee should remove combatant
        self._run_actions(action_dict)
        # this ends combat, so combathandler should be gone
        self.assertIsNone(self.combathandler.pk)


class TestEvAdventureTwitchCombatHandler(EvenniaCommandTestMixin, _CombatTestBase):
    def setUp(self):
        super().setUp()

        # in order to use the EvenniaCommandTestMixin we need these variables defined
        self.char1 = self.combatant
        self.account = None

        self.combatant_combathandler = (
            combat_twitch.EvAdventureCombatTwitchHandler.get_or_create_combathandler(
                self.combatant, key="combathandler"
            )
        )
        self.target_combathandler = (
            combat_twitch.EvAdventureCombatTwitchHandler.get_or_create_combathandler(
                self.target, key="combathandler"
            )
        )

    def test_get_sides(self):
        sides = self.combatant_combathandler.get_sides(self.combatant)
        self.assertEqual(sides, ([self.combatant], [self.target]))

    def test_give_advantage(self):
        self.combatant_combathandler.give_advantage(self.combatant, self.target)
        self.assertTrue(self.combatant_combathandler.advantage_against[self.target])

    def test_give_disadvantage(self):
        self.combatant_combathandler.give_disadvantage(self.combatant, self.target)
        self.assertTrue(self.combatant_combathandler.disadvantage_against[self.target])

    @patch("evennia.contrib.tutorials.evadventure.combat_twitch.unrepeat", new=Mock())
    @patch("evennia.contrib.tutorials.evadventure.combat_twitch.repeat", new=Mock(return_value=999))
    def test_queue_action(self):
        """Test so the queue action cleans up tickerhandler correctly"""

        actiondict = {"key": "hold"}
        self.combatant_combathandler.queue_action(actiondict)

        self.assertIsNone(self.combatant_combathandler.current_ticker_ref)

        actiondict = {"key": "hold", "dt": 5}
        self.combatant_combathandler.queue_action(actiondict)
        self.assertEqual(self.combatant_combathandler.current_ticker_ref, 999)

    @patch("evennia.contrib.tutorials.evadventure.combat_twitch.unrepeat", new=Mock())
    @patch("evennia.contrib.tutorials.evadventure.combat_twitch.repeat", new=Mock())
    def test_execute_next_action(self):
        self.combatant_combathandler.action_dict = {
            "key": "hold",
            "dummy": "foo",
            "repeat": False,
        }  # to separate from fallback

        self.combatant_combathandler.execute_next_action()
        # should now be back to fallback
        self.assertEqual(
            self.combatant_combathandler.action_dict,
            self.combatant_combathandler.fallback_action_dict,
        )

    @patch("evennia.contrib.tutorials.evadventure.combat_twitch.unrepeat", new=Mock())
    def test_check_stop_combat(self):
        """Test combat-stop functionality"""

        # noone remains (both combatant/target <0 hp
        # get_sides does not include the caller
        self.combatant_combathandler.get_sides = Mock(return_value=([], []))
        self.combatant_combathandler.stop_combat = Mock()

        self.combatant.hp = -1
        self.target.hp = -1

        self.combatant_combathandler.check_stop_combat()
        self.combatant.msg.assert_called_with(
            text=("Noone stands after the dust settles.", {}), from_obj=self.combatant
        )
        self.combatant_combathandler.stop_combat.assert_called()

        # only one side wiped out
        self.combatant.hp = 10
        self.target.hp = -1
        self.combatant_combathandler.get_sides = Mock(return_value=([self.combatant], []))
        self.combatant_combathandler.check_stop_combat()
        self.combatant.msg.assert_called_with(
            text=("The combat is over.", {}), from_obj=self.combatant
        )

    @patch("evennia.contrib.tutorials.evadventure.combat_twitch.unrepeat", new=Mock())
    @patch("evennia.contrib.tutorials.evadventure.combat_twitch.repeat", new=Mock())
    def test_hold(self):
        self.call(combat_twitch.CmdHold(), "", "You hold back, doing nothing")
        self.assertEqual(self.combatant_combathandler.action_dict, {"key": "hold"})

    @patch("evennia.contrib.tutorials.evadventure.combat_twitch.unrepeat", new=Mock())
    @patch("evennia.contrib.tutorials.evadventure.combat_twitch.repeat", new=Mock())
    def test_attack(self):
        """Test attack action in the twitch combathandler"""
        self.call(combat_twitch.CmdAttack(), self.target.key, "You attack testmonster!")
        self.assertEqual(
            self.combatant_combathandler.action_dict,
            {"key": "attack", "target": self.target, "dt": 3, "repeat": True},
        )

    @patch("evennia.contrib.tutorials.evadventure.combat_twitch.unrepeat", new=Mock())
    @patch("evennia.contrib.tutorials.evadventure.combat_twitch.repeat", new=Mock())
    def test_stunt(self):
        boost_result = {
            "key": "stunt",
            "recipient": self.combatant,
            "target": self.target,
            "advantage": True,
            "stunt_type": Ability.STR,
            "defense_type": Ability.STR,
            "dt": 3,
        }
        foil_result = {
            "key": "stunt",
            "recipient": self.target,
            "target": self.combatant,
            "advantage": False,
            "stunt_type": Ability.STR,
            "defense_type": Ability.STR,
            "dt": 3,
        }

        self.call(
            combat_twitch.CmdStunt(),
            f"STR {self.target.key}",
            "You prepare a stunt!",
            cmdstring="boost",
        )
        self.assertEqual(self.combatant_combathandler.action_dict, boost_result)

        self.call(
            combat_twitch.CmdStunt(),
            f"STR me {self.target.key}",
            "You prepare a stunt!",
            cmdstring="boost",
        )
        self.assertEqual(self.combatant_combathandler.action_dict, boost_result)

        self.call(
            combat_twitch.CmdStunt(),
            f"STR {self.target.key}",
            "You prepare a stunt!",
            cmdstring="foil",
        )
        self.assertEqual(self.combatant_combathandler.action_dict, foil_result)

        self.call(
            combat_twitch.CmdStunt(),
            f"STR {self.target.key} me",
            "You prepare a stunt!",
            cmdstring="foil",
        )
        self.assertEqual(self.combatant_combathandler.action_dict, foil_result)

    @patch("evennia.contrib.tutorials.evadventure.combat_twitch.unrepeat", new=Mock())
    @patch("evennia.contrib.tutorials.evadventure.combat_twitch.repeat", new=Mock())
    def test_useitem(self):
        item = create.create_object(
            EvAdventureConsumable, key="potion", attributes=[("uses", 2)], location=self.combatant
        )

        self.call(combat_twitch.CmdUseItem(), "potion", "You prepare to use potion!")
        self.assertEqual(
            self.combatant_combathandler.action_dict,
            {"key": "use", "item": item, "target": self.combatant, "dt": 3},
        )

        self.call(
            combat_twitch.CmdUseItem(),
            f"potion on {self.target.key}",
            "You prepare to use potion!",
        )
        self.assertEqual(
            self.combatant_combathandler.action_dict,
            {"key": "use", "item": item, "target": self.target, "dt": 3},
        )

    @patch("evennia.contrib.tutorials.evadventure.combat_twitch.unrepeat", new=Mock())
    @patch("evennia.contrib.tutorials.evadventure.combat_twitch.repeat", new=Mock())
    def test_wield(self):
        sword = create.create_object(EvAdventureWeapon, key="sword", location=self.combatant)
        runestone = create.create_object(
            EvAdventureWeapon, key="runestone", location=self.combatant
        )

        self.call(combat_twitch.CmdWield(), "sword", "You reach for sword!")
        self.assertEqual(
            self.combatant_combathandler.action_dict, {"key": "wield", "item": sword, "dt": 3}
        )

        self.call(combat_twitch.CmdWield(), "runestone", "You reach for runestone!")
        self.assertEqual(
            self.combatant_combathandler.action_dict, {"key": "wield", "item": runestone, "dt": 3}
        )
